"""PASTA Implementation"""
import torch
import abc, json
from contextlib import contextmanager
from functools import partial
from pathlib import Path
from typing import Any, Literal, Optional, Sequence, cast, overload, Tuple

import transformers 
from pastalib.utils import tokenizer_utils
from pastalib.utils.typing import (
    Model,
    Dataset,
    Device,
    ModelInput,
    ModelOutput,
    StrSequence,
    Tokenizer,
    TokenizerOffsetMapping,
)


class PASTA(abc.ABC):
    """
    Create PASTA to steer attentions of transformer models. 

    Args:
        model ([`transformers.PreTrainedModel`]): The model to be steered. 
        tokenizer ([`transformers.PreTrainedTokenizer`]): The model's tokenizer. 
        head_config (`dict`): The config to control which attention heads to be steered. 
        alpha (`float`): The scaling coefficient of attention steering. 
        scale_position (`str`): To upweight the scores of highlighted tokens (`include`), 
            or to downweight those of unselected tokens (`exclude`). 

    Returns:
        PASTA: PASTA steerer that can register the pre-forward hooks on target models. 

    """

    ATTN_MODULE_NAME = {
        "gptj": "transformer.h.{}.attn",
        "llama": "model.layers.{}.self_attn",
        "mistral": "model.layers.{}.self_attn",
        "gemma": "model.layers.{}.self_attn",
        "phi3mini": "model.layers.{}.self_attn"
    }
    ATTENTION_MASK_ARGIDX = {
        "gptj": 2, 
        "llama": 1, 
        "mistral": 1, 
        "gemma": 1,
    }
    def __init__(
        self, 
        model: Model, 
        tokenizer: Tokenizer, 
        head_config: dict|list|None = None, 
        alpha: float = 0.01, 
        scale_position: str = "exclude", 
    ):
        self.model = model
        self.tokenizer = tokenizer
        self.setup_model(model)

        self.alpha = alpha
        self.scale_position = scale_position
        self.setup_head_config(head_config)

        assert self.scale_position in ['include', 'exclude', 'generation']
        assert self.alpha > 0

    def setup_model(self, model):
        """Obtain the model type and complete the configuration."""
        if isinstance(model, transformers.LlamaForCausalLM):
            self.model_name = "llama"
            self.num_attn_head = model.config.num_attention_heads
        elif isinstance(model, transformers.GPTJForCausalLM):
            self.model_name = "gptj"
            self.num_attn_head = model.config.n_head
        elif isinstance(model, transformers.MistralForCausalLM):
            self.model_name = "mistral"
            self.num_attn_head = model.config.num_attention_heads
        elif isinstance(model, transformers.GemmaForCausalLM):
            self.model_name = "gemma"
            self.num_attn_head = model.config.num_attention_heads
        elif model.__class__.__name__ == "Phi3ForCausalLM":
            self.model_name = "phi3"
        else:
            raise ValueError("Unimplemented Model Type.")
        
    def setup_head_config(self, head_config):
        """
        Config the attention heads to be steered.

        If `head_config` is `list` of layer index, PASTA will steer the entire layers. 
        """
        if isinstance(head_config, dict):
            self.head_config = {int(k):v for k,v in head_config.items()} 
            self.all_layers_idx = [int(key) for key in head_config]
        elif isinstance(head_config, list):
            self.all_layers_idx = [int(v) for v in head_config]
            self.head_config = {
                idx:list(range(self.num_attn_head)) for idx in self.all_layers_idx
            }
        else:
            raise ValueError(f"Incorrect head config: {head_config}")
    
    def _maybe_batch(self, text: str | StrSequence) -> StrSequence:
        """Batch the text if it is not already batched."""
        if isinstance(text, str):
            return [text]
        return text

    def token_ranges_from_batch(
        self,
        strings: str | StrSequence,
        substrings: str | StrSequence,
        offsets_mapping: Sequence[TokenizerOffsetMapping],
        occurrence: int = 0,
    ) -> torch.Tensor:
        """Return shape (batch_size, 2) tensor of token ranges for (str, substr) pairs."""
        strings = self._maybe_batch(strings)
        substrings = self._maybe_batch(substrings)
        if len(strings) != len(substrings):
            raise ValueError(
                f"got {len(strings)} strings but only {len(substrings)} substrings"
            )
        return torch.tensor(
            [
                tokenizer_utils.find_token_range(
                    string, substring, offset_mapping=offset_mapping, occurrence=occurrence
                )
                for string, substring, offset_mapping in zip(
                    strings, substrings, offsets_mapping
                )
            ]
        )

    def edit_attention_mask(
        self, 
        module: torch.nn.Module, 
        input_args: tuple,
        input_kwargs: dict, 
        head_idx: list[int],
        token_range: torch.Tensor, 
        input_len: int, 
    ):
        """
        The hook function registerred pre-forward for attention models. 

        Args: 
            module ([`torch.nn.Module`]): The registerred attention modules. 
            input_args (`tuple`): The positional arguments of forward function. 
            input_kwargs (`dict`): The keyword arguments of forward function. 
            head_idx (`list[int]`): The index of heads to be steered. 
            token_range (`torch.Tensor`): A B*2 tensor, 
                suggesting the index range of hightlight tokens.  
            input_len (`int`): The length L of inputs.

        Returns: 
            tuple, dict: return the modified `attention_mask`,
                while not changing other input arguments. 
        """
        if "attention_mask" in input_kwargs:
            attention_mask = input_kwargs['attention_mask'].clone()
        elif input_args is not None:
            arg_idx = self.ATTENTION_MASK_ARGIDX[self.model_name]
            attention_mask = input_args[arg_idx].clone()
        else:
            raise ValueError(f"Not found attention masks in {str(module)}")
        
        bsz, head_dim, tgt_len, src_len = attention_mask.size()
        dtype, device = attention_mask.dtype, attention_mask.device
        if head_dim != self.num_attn_head:
            attention_mask = attention_mask.expand(
                bsz, self.num_attn_head, tgt_len, src_len
            ).clone()
        scale_constant = torch.Tensor([self.alpha]).to(dtype).to(device).log()
        
        for bi, (ti,tj) in enumerate(token_range.tolist()):
            if self.scale_position == "include":
                attention_mask[bi, head_idx, :, ti:tj] += scale_constant
            else:
                attention_mask[bi, head_idx, :, :ti] += scale_constant
                attention_mask[bi, head_idx, :, tj:input_len] += scale_constant
        
        if self.model_name in ["llama", "mistral", "gemma", "phi3"]:
            attention_mask.old_size = attention_mask.size 
            attention_mask.size = lambda:(bsz, 1, tgt_len, src_len)
        
        if "attention_mask" in input_kwargs:
            input_kwargs['attention_mask'] = attention_mask 
            return input_args, input_kwargs
        else:
            return (input_args[:arg_idx], attention_mask, *input_args[arg_idx+1:]), input_kwargs

    def edit_multisection_attention(
        self, 
        module: torch.nn.Module, 
        input_args: tuple,
        input_kwargs: dict, 
        head_idx: list[int],
        token_ranges: list[torch.Tensor], 
        input_len: int, 
    ):
        """
        The hook function registerred pre-forward for attention models. 

        Args: 
            module ([`torch.nn.Module`]): The registerred attention modules. 
            input_args (`tuple`): The positional arguments of forward function. 
            input_kwargs (`dict`): The keyword arguments of forward function. 
            head_idx (`list[int]`): The index of heads to be steered. 
            token_ranges (`torch.Tensor`): A list of B*2 tensors, 
                suggesting the index range of hightlight tokens of multiple sections.  
            input_len (`int`): The length L of inputs.

        Returns: 
            tuple, dict: return the modified `attention_mask`,
                while not changing other input arguments. 
        """
        if "attention_mask" in input_kwargs:
            attention_mask = input_kwargs['attention_mask'].clone()
        elif input_args is not None:
            arg_idx = self.ATTENTION_MASK_ARGIDX[self.model_name]
            attention_mask = input_args[arg_idx].clone()
        else:
            raise ValueError(f"Not found attention masks in {str(module)}")
        
        bsz, head_dim, tgt_len, src_len = attention_mask.size()
        dtype, device = attention_mask.dtype, attention_mask.device
        if head_dim != self.num_attn_head:
            attention_mask = attention_mask.expand(
                bsz, self.num_attn_head, tgt_len, src_len
            ).clone()
        scale_constant = torch.Tensor([self.alpha]).to(dtype).to(device).log()
        
        for token_range in token_ranges:
            for bi, (ti,tj) in enumerate(token_range.tolist()):
                if self.scale_position == "include":
                    attention_mask[bi, head_idx, :, ti:tj] += scale_constant
                elif self.scale_position == "exclude":
                    attention_mask[bi, head_idx, :, :ti] += scale_constant
                    attention_mask[bi, head_idx, :, tj:input_len] += scale_constant
                elif self.scale_position == "generation":
                    attention_mask[bi, head_idx, :, :input_len] += scale_constant 
                else:
                    raise ValueError(f"Unexcepted {self.scale_position}.")
        if self.scale_position == "include":
            attention_mask[:, head_idx, :, :input_len] -= scale_constant
        
        if self.model_name in ["llama", "mistral"]:
            attention_mask.old_size = attention_mask.size 
            attention_mask.size = lambda:(bsz, 1, tgt_len, src_len)
        
        if "attention_mask" in input_kwargs:
            input_kwargs['attention_mask'] = attention_mask 
            return input_args, input_kwargs
        else:
            return (input_args[:arg_idx], attention_mask, input_args[arg_idx+1:]), input_kwargs


    @contextmanager
    def apply_steering(
        self, 
        model: Model, 
        strings: list, 
        substrings: list, 
        model_input: ModelInput, 
        offsets_mapping: Sequence[TokenizerOffsetMapping], 
        occurrence: int = 0,
    ):
        """
        The function of context manager to register the pre-forward hook on `model`. 

        Args:
            model ([`transformers.PreTrainedModel`]): The transformer model to be steered. 
            strings (`list[str]`): The input strings. 
            substrings (`list[list[str]]` or list[str]): The highlighted input spans for each string. 
            model_input (`transformers.BatchEncoding`): The batched model inputs. 
            offsets_mapping (`TokenizerOffsetMapping`): The offset mapping outputed by
                the tokenizer when encoding the `strings`. 
        """
        if isinstance(substrings[0], str):
            substrings = [substrings]

        token_ranges = []
        for sections in substrings:
            token_range = self.token_ranges_from_batch(
                strings, sections, offsets_mapping, occurrence=occurrence,
            )
            token_ranges.append(token_range)

        registered_hooks = []
        for layer_idx in self.all_layers_idx:
            name = self.ATTN_MODULE_NAME[self.model_name].format(layer_idx)
            module = model.get_submodule(name)
            # Prepare the hook function with partial arguments being fixed. 
            # Pass the head_idx, token_range, input_len for each attention module in advance. 
            hook_func = partial(
                self.edit_multisection_attention, 
                head_idx = self.head_config[layer_idx],
                token_ranges = token_ranges, 
                input_len = model_input['input_ids'].size(-1)
            )
            registered_hook = module.register_forward_pre_hook(hook_func, with_kwargs=True)
            registered_hooks.append(registered_hook)
        try:
            yield model
        except Exception as error:
            raise error
        finally:
            for registered_hook in registered_hooks:
                registered_hook.remove()
    

    def inputs_from_batch(
        self, 
        text: str | StrSequence,
        tokenizer: Tokenizer|None = None,
        device: Optional[Device] = None,
    ) -> tuple[ModelInput, Sequence[TokenizerOffsetMapping]]:
        """Precompute model inputs."""
        if tokenizer is None:
            tokenizer = self.tokenizer
        with tokenizer_utils.set_padding_side(tokenizer, padding_side="left"):
            inputs = tokenizer(
                text,
                return_tensors="pt",
                truncation=False,
                padding="longest",
                return_offsets_mapping=True,
            )
            offset_mapping = inputs.pop("offset_mapping")
        if device is not None:
            inputs = inputs.to(device)
        return inputs, offset_mapping

    @classmethod
    def load_head_config(cls, file:str|Path):
        """Load the `head_config` from JSON file."""
        with open(file, "r") as f:
            head_config = json.load(f)
        return head_config 




"""Metrics used for evaluating the editors.

Lots of this code is taken from the ROME CounterFact evaluation, for which the source
code can be found: https://github.com/kmeng01/rome/blob/main/experiments/py/eval_utils_counterfact.py
"""
import logging
from dataclasses import dataclass
from typing import Any, Sequence

from evaluation.utils.typing import ArrayLike, StrSequence

import nltk
import numpy as np
from dataclasses_json import DataClassJsonMixin
from sklearn.feature_extraction.text import TfidfVectorizer
from tqdm.auto import tqdm

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Metric(DataClassJsonMixin):
    """An aggregate metric."""

    mean: float
    std: float
    values: ArrayLike | None = None

    def without_values(self) -> "Metric":
        """Return the metric without the values stored."""
        return Metric(mean=self.mean, std=self.std)

    @staticmethod
    def aggregate(values: ArrayLike, store_values: bool = True) -> "Metric":
        """Aggregate mean/std of the values."""
        return Metric(
            np.mean(values), np.std(values), values=values if store_values else None
        )


@dataclass(frozen=True)
class EfficacyMetrics(DataClassJsonMixin):
    """Efficacy metrics."""

    score: Metric
    magnitude: Metric

    def without_values(self) -> "EfficacyMetrics":
        """Return the metrics without the values stored."""
        return EfficacyMetrics(
            score=self.score.without_values(),
            magnitude=self.magnitude.without_values(),
        )


def efficacy(
    p_targets: Sequence[ArrayLike],
    p_comparators: Sequence[ArrayLike],
    assume_log_probs: bool = True,
    store_values: bool = True,
) -> EfficacyMetrics:
    """Compute efficacy on metrics.

    Efficacy is determined by a score and a magnitude. The score is how often
    p(target) > p(comparator). The magnitude is the average p(target) - p(comparator).

    Inputs are two sequences. Each element should be one or more measurements of
    the probability (for e.g. different prompts). This function will first average
    across those inner lists, then average across the whole list.
    """
    _validate_same_length(p_targets=p_targets, p_comparators=p_comparators)

    scores, magnitudes = [], []
    for i, (p_target, p_comparator) in enumerate(zip(p_targets, p_comparators)):
        _validate_same_length(
            **{f"p_target_{i}": p_target, f"p_comparator_{i}": p_comparator}
        )

        if assume_log_probs:
            p_target = np.exp(p_target)
            p_comparator = np.exp(p_comparator)
        else:
            p_target = np.array(p_target)
            p_comparator = np.array(p_comparator)

        score = np.mean(p_target > p_comparator)
        scores.append(score)

        magnitude = np.mean(p_target - p_comparator)
        magnitudes.append(magnitude)

    return EfficacyMetrics(
        score=Metric.aggregate(scores, store_values=store_values),
        magnitude=Metric.aggregate(magnitudes, store_values=store_values),
    )


def average_tfidf_similarity(
    generations: Sequence[StrSequence],
    references: Sequence[StrSequence],
    tfidf_vectorizer: TfidfVectorizer,
    desc: str | None = None,
) -> Metric:
    """Compute TF-IDF similarity between generated text and reference texts.

    This is used downstream to compute the "consistency" and "essence" scores.
    """
    _validate_same_length(generations=generations, references=references)
    if desc is None:
        desc = "tfidf similarity"
    similarities = [
        tfidf_similarity(gs, rs, tfidf_vectorizer)
        for gs, rs in tqdm(list(zip(generations, references)), desc=desc)
    ]
    return Metric.aggregate(similarities)


def tfidf_similarity(
    source: str | StrSequence,
    reference: str | StrSequence,
    tfidf_vectorizer: TfidfVectorizer,
) -> float:
    """Return TfIdf similarity between the texts."""
    if isinstance(source, str):
        source = [source]
    if isinstance(reference, str):
        reference = [reference]
    sv, rv = tfidf_vectorizer.transform([" ".join(source), " ".join(reference)]).A
    return vector_similarity(sv, rv)


def vector_similarity(a: np.ndarray, b: np.ndarray, eps: float = 1e-6) -> float:
    """Compute cosine similarity between two vectors."""
    return np.dot(a, b) / (np.linalg.norm(a) + eps) / (np.linalg.norm(b) + eps)


def average_weighted_n_gram_entropy(
    generations: Sequence[StrSequence], desc: str | None = None, **kwargs: Any
) -> Metric:
    """Compute fluency score.

    Args:
        generations: Any number of generated strings from the model.

    Returns:
        Average n-gram entropy across generations. Can specify multiple ns by
        setting the ns= kwarg. By default, uses 2-grams and 3-grams.

    """
    if desc is None:
        desc = "weighed n-gram entropy"
    entropies = [
        np.mean([weighted_n_gram_entropy(text, **kwargs) for text in texts])
        for texts in tqdm(generations, desc=desc)
    ]
    return Metric.aggregate(entropies)


def weighted_n_gram_entropy(
    texts: str | StrSequence,
    ns: Sequence[int] = (2, 3),
    weights: Sequence[float] = (2 / 3, 4 / 3),
) -> float:
    """Return weighted n-gram entropy for different values of n."""
    _validate_same_length(ns=ns, weights=weights)
    if isinstance(texts, str):
        texts = [texts]
    entropies = []
    for text in texts:
        entropies_by_n = np.array([n_gram_entropy(text, n) for n in ns])
        entropy = np.mean(entropies_by_n * np.array(weights))
        entropies.append(entropy)
    return np.mean(entropies)


def n_gram_entropy(text: str, n: int) -> float:
    """Return entropy of n-gram distribution in text."""
    counts = _n_gram_counts(text, n)
    dist = np.array([count for _, count in counts.items()], dtype=np.float32)
    dist /= dist.sum()
    entropy = np.sum(-dist * np.log(dist) / np.log(2))
    return entropy.item()


def _validate_same_length(**kwargs: Sequence | ArrayLike) -> None:
    """Validate all batch sizes are the same."""
    lengths = {key: len(seq) for key, seq in kwargs.items()}
    if len(set(lengths.values())) > 1:
        message = f"inconsistent batch sizes:" + "\n\t"
        message += "\n\t".join(f"{key}={length}" for key, length in lengths.items())
        raise ValueError(message)


def _n_gram_counts(text: str, n: int) -> dict[tuple[str, ...], int]:
    """Return the n-gram counts for the text."""
    tokens = nltk.word_tokenize(text)
    ngrams = nltk.ngrams(tokens, n)
    return nltk.FreqDist(ngrams)



import logging
import random
import json 
from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from itertools import chain, combinations
from typing import Any, Callable, Sequence, cast

from evaluation import data, metrics, models, precompute
from evaluation.utils import experiment_utils
from evaluation.utils.typing import Dataset, Device, StrSequence

import numpy as np
import scipy.stats
import spacy
import torch
import torch.utils.data
from dataclasses_json import DataClassJsonMixin
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics import accuracy_score, f1_score, matthews_corrcoef
from tqdm.auto import tqdm


@dataclass(frozen=True)
class InstructionEvaluationSample:
    """Wrapper around format evaluation sample."""

    id: str
    prompt: str
    generation: str

    predictions: list[str]
    target: str

    logp_predictions: list[float]
    logp_target: float

    fluency: float
    consistency: float

    instruction_evaluation: dict 
    sample_attn_scores: dict 


@dataclass(frozen=True)
class InstructionEvaluationMetrics(DataClassJsonMixin):
    """Wrapper around aggregated error correction metrics."""

    top1_accuracy: float
    topk_accuracy: float
    k: int

    fluency: metrics.Metric
    consistency: metrics.Metric

    instruction_evaluation: dict 


@dataclass(frozen=True)
class BiosBiasInstructionEvaluationResults(DataClassJsonMixin):
    """Wrapper around error correction benchmark."""

    samples: list[InstructionEvaluationSample]
    metrics: InstructionEvaluationMetrics 
    attentions: dict 


def get_all_substring_index(string, substring):
    index = [i for i in range(len(string)) if string.startswith(substring, i)]
    return index


class InstructionEvaluator(object):
    """
    The Evaluator for instruction following tasks: JSON Formatting and Pronouns Changing. 

    Args:
        task (`str`): The name of evaluation task (`json` or `pronchange`). 
        prompt_idx (`int`): The index of the prompte template. See `DEFAULT_PROMPT`. 

    Returns: 
        InstructionEvaluator
    """
    INSTRUCTION_TASK_NAMES = [
        "json", "pronchange", 
    ]
    DEFAULT_PROMPT = {
        "json": [
            '''
            Answer the occupation of {} and generate the answer as json format. Here is an example: \{{'name': , 'occupation': ,\}}. Now generate the answer:
            ''',
            '''
            Generate the occupation of {} as json format: 
            ''',
            '''
            Answer the occupation of {} and generate the answer as json format: 
            ''',
            '''
            **Answer the occupation of {} and generate the answer as json format. Here is an example: \{{'name': , 'occupation': ,\}}. Now generate the answer:**
            ''',
            '''
            "Answer the occupation of {} and generate the answer as json format. Here is an example: \{{'name': , 'occupation': ,\}}. Now generate the answer:"
            ''',
            '''
            Answer the occupation of {} and generate the answer as json format. Here is an example: \{{"name": , "occupation": ,\}}. 
            ''',
        ],
        "pronchange": [
            '''
            For the aforementioned text, substitute 'she' and 'he' with 'they' and generate the occupation of {} after changing pronouns:
            ''',
            '''
            For the aforementioned descriptions, replace 'she' and 'he' with 'they' in the aformentioned text and generate the new text after replacing the pronouns:
            ''',
            '''
            Change 'she' and 'he' with 'they' and answer the occupation of {} after replacing the pronouns: 
            ''',
            '''
            **For the aforementioned text, substitute 'she' and 'he' with 'they' and generate the occupation of {} after changing pronouns:**
            ''',
            '''
            "For the aforementioned text, substitute 'she' and 'he' with 'they' and generate the occupation of {} after changing pronouns:"
            ''',
        ], 
    }
    JSON_ALIKE_CHARS = ["{", "}", ":", ","]
    PRON_CHANGE_TARGET = ["she", "he"] 
    PRON_ALL_CHANGE_TARGET = ["she", "he", "her", "him", "hers", "his", "herself", "himself"] 

    def __init__(
        self, 
        task:str,
        prompt_idx: int|list|None = None, 
    ):
        self.task = task
        assert task in self.INSTRUCTION_TASK_NAMES 

        if isinstance(prompt_idx, int): 
            self.prompt_temple = [self.DEFAULT_PROMPT[task][prompt_idx]]
            self.prompt_idx = [prompt_idx]
        elif isinstance(prompt_idx, list): 
            self.prompt_temple = [self.DEFAULT_PROMPT[task][i] for i in prompt_idx]
            self.prompt_idx = prompt_idx 
        else:
            self.prompt_temple = self.DEFAULT_PROMPT[task]
            self.prompt_idx = list(range(len(self.DEFAULT_PROMPT[task])))
        
        self.prompt_temple = [prompt.strip() for prompt in self.prompt_temple] 
        self.num_prompt = len(self.prompt_temple) 
        self.nlp = spacy.load("en_core_web_sm")
        
    def parapare_prompt_inputs(
            self, 
            contexts: list, 
            entities: list, 
            other_text_inputs: tuple[list], 
        ):
        """Apply the prompt templates and generation input strings for models."""
        prompts = [
            self.apply_prompt_templete(context, entity, prompt, idx)
                for context,entity in zip(contexts, entities) for prompt,idx in zip(self.prompt_temple, self.prompt_idx)
        ]
        instructions = [
            prompt.format(entity) for entity in entities for prompt in self.prompt_temple 
        ]
        new_entities = [
            entity for entity in entities for i in range(self.num_prompt)
        ]
        new_other_inputs = []
        for text_input in other_text_inputs:
            new_text_input = [
                val for val in text_input for i in range(len(self.prompt_temple))
            ]
            new_other_inputs.append(new_text_input)
        
        return prompts, instructions, new_entities, tuple(new_other_inputs)
    
    def evaluate_sample(self, generation, target=None, **kwargs):
        result = {}
        if self.task == "json":
            result["json_EM"] = self.evaluate_json_EM(generation)
            if target is not None:
                result['json_pred_acc'] = self.evaluate_json_pred(generation, target)
            for requrie_num in range(2, len(self.JSON_ALIKE_CHARS)+1):
                result[f"json_alike_{requrie_num}"] = self.evaluate_json_alike(generation, requrie_num)
            return result 
        elif self.task == "pronchange":
            result["pron_sub_acc"], result["pron_all_acc"] = self.evaluate_pronounce_change(generation)
        return result 
    
    def apply_prompt_templete(self, context, entity, prompt_templete, prompt_idx):
        """Apply the prompt template"""
        prompt = context+'\n'+prompt_templete.format(entity)+'\n' 
        if self.task == "json" and prompt_idx in [5]:
            prompt = prompt_templete.format(entity)+'\n'+context+'\n'+"Now generate the answer:"
        else:
            prompt = context+'\n'+prompt_templete.format(entity)+'\n' 
        return prompt
    
    def evaluate_json_alike(self, generation, require_num):
        """
        Evaluate how generations are alike JSON format
        by checking whether generations include '{', ':', ',', '}'. 
        """
        all_combs = list(combinations(self.JSON_ALIKE_CHARS, require_num))
        for chars in all_combs:
            if all([char in generation for char in chars]):
                return True 
        return False 

    def evaluate_json_EM(self, generation):
        """
        Evaluate whether generations exactly match JSON format 
        by `json.loads` function. 
        """
        generation = generation.strip()
        if "{" not in generation and "}" not in generation:
            return False
        
        string_begs = get_all_substring_index(generation, "{")
        string_ends = get_all_substring_index(generation, "}")
        # Try to load every substring starting from '{' and end with '}'
        for idx_beg in string_begs:
            for idx_end in string_ends:
                substring = generation[idx_beg:idx_end+1]
                try:
                    json.loads(substring)
                    return True 
                except ValueError:
                    continue
        return False 
    
    def evaluate_json_pred(self, generation, target):
        """
        Evaluate whether generated JSON outputs correctly include 
        the traget label by checking its `values` after `json.loads`
        """
        generation = generation.strip()
        if "{" not in generation and "}" not in generation:
            return False
        
        string_begs = get_all_substring_index(generation, "{")
        string_ends = get_all_substring_index(generation, "}")
        for idx_beg in string_begs:
            for idx_end in string_ends:
                substring = generation[idx_beg:idx_end+1]
                try:
                    pred = json.loads(substring)
                    pred_values = [p.lower() if isinstance(p, str) else p for p in pred.values()]
                    if target.lower() in pred_values:
                        return True 
                    for value in pred_values:
                        if isinstance(value, str) and target.lower() in value:
                            return True
                except ValueError:
                    continue
        return False 
    
    def evaluate_pronounce_change(self, generation):
        """
        Evaluate the accuracy of successfully changing the pronouns 
        """
        generation = generation.lower()
        all_tokens = [token.text for token in self.nlp(generation)]
        pron_sub, pron_all = True, True
        for word in all_tokens:
            if word in self.PRON_CHANGE_TARGET:
                pron_sub = False
            if word in self.PRON_ALL_CHANGE_TARGET:
                pron_all = False
        return pron_sub, pron_all
        
    def aggregate_evaluation_results(self, samples):
        """Aggergate the evaluation results"""
        result_key = list(samples[0].instruction_evaluation.keys())
        grouped_results = {}
        for sample in samples:
            if sample.id not in grouped_results:
                grouped_results[sample.id] = {key:[] for key in result_key}
            for key in sample.instruction_evaluation:
                grouped_results[sample.id][key].append(sample.instruction_evaluation[key])
        
        all_result = {key:[] for key in result_key}
        for _, sample_eval in grouped_results.items():
            for key in sample_eval:
                all_result[key].append(any(sample_eval[key])) 

        return {key:sum(value)/len(value) for key,value in all_result.items()}
    
    def is_pred_true_sample_result(self, sample):
        if self.task == "json":
            return sample.instruction_evaluation["json_EM"]
        elif self.task == "pronchange":
            return sample.instruction_evaluation["pron_sub_acc"]

    def get_generation_label(self, context, entity, target):
        if self.task == "json":
            label = '{"name":"%s", "occupation":"%s"}'%(entity, target)
        elif self.task == "pronchange":
            label = context.lower().replace('she', 'they').replace(
                'he', 'they').replace('him', 'them').replace('her', 'them')
        return label 

    def prepare_fewshot_examples(
            self, fewshot_samples, example_sep = "\n###\n", text_target_sep = " ", 
        ):
        """Prepare few-shot demonstration"""
        labeled_example = [] 
        for example in fewshot_samples:
            context = example['context']
            entity = example['entity']
            target = example['target_mediated'] 

            label = self.get_generation_label(context, entity, target) 
            labeled_example.append(context+text_target_sep+label)
        
        fewshot_examples = example_sep.join(labeled_example) + example_sep
        return fewshot_examples 


# For attention score recodring 
def concatenate_attention_scores(output_attentions):
    num_layer = len(output_attentions[0]) 
    num_new_tokens = len(output_attentions) 
    all_attn_scores = [] 
    for layer_id in range(num_layer):
        token_len = output_attentions[0][layer_id].size(-1)
        layer_attn_score = [output_attentions[0][layer_id][:,:,-1,:].unsqueeze(2)]
        for tid in range(1, num_new_tokens):
            layer_attn_score.append(
                output_attentions[tid][layer_id][:,:,0,:token_len].unsqueeze(2)
            )
        all_attn_scores.append(torch.cat(layer_attn_score, dim=2))
    return all_attn_scores 


def calcluate_attn_metric(sub_attn_score, metric):
    if metric == "token_mean_head_all":
        return sub_attn_score.mean(-1).tolist()
    elif metric == "token_head_mean":
        return sub_attn_score.mean().tolist()
    elif metric == "token_head_median":
        return sub_attn_score.float().median(-1).values.median().item()
    elif metric == "token_mean_head_median":
        return sub_attn_score.float().mean(-1).median().item()
    elif metric == "token_mean_head_std":
        return sub_attn_score.mean(-1).std().item() 
    else:
        raise ValueError(f"Unimplement the metric {metric}")


def process_sample_attention_score(
        all_attn_scores, tr_dict, idx, 
        sections=["first", "all"],
        metrics=["token_head_median"]
    ):
    sample_attn_scores = {sec: {key:{metric:[] for metric in metrics} for key in tr_dict} for sec in sections}
    for layer_id in range(len(all_attn_scores)):
        for section in sections:
            if section == "first":
                attn_score = all_attn_scores[layer_id][idx,:,0,:]
            else:
                attn_score = all_attn_scores[layer_id][idx,:,:,:].mean(1)
            for key in sample_attn_scores[section]:
                for metric in metrics:
                    assert tr_dict[key][0]<tr_dict[key][1]
                    sample_attn_scores[section][key][metric].append(
                        calcluate_attn_metric(attn_score[:,tr_dict[key][0]:tr_dict[key][1]], metric)
                    )
                sample_attn_scores[section][key]["token_range"] = tr_dict[key] 
    return sample_attn_scores


def average_sample_attention_scores(samples):
    attn_scores = {}
    for sample in samples: 
        sample_attn_scores = sample.sample_attn_scores
        for sec in sample_attn_scores:
            if sec not in attn_scores:
                attn_scores[sec] = {}
            for key in sample_attn_scores[sec]:
                if key not in attn_scores[sec]:
                    attn_scores[sec][key] = {}
                for metric in sample_attn_scores[sec][key]:
                    if metric == "token_range":
                        continue
                    if metric not in attn_scores[sec][key]:
                        attn_scores[sec][key][metric] = []
                    attn_scores[sec][key][metric].append(
                        sample_attn_scores[sec][key][metric]
                    )
    for sec in attn_scores:
        for key in attn_scores[sec]:
            for metric in attn_scores[sec][key]:
                attn_scores[sec][key][metric] = torch.tensor(
                    attn_scores[sec][key][metric]
                ).mean(0).tolist()
    return attn_scores 


def average_sample_attention_scores_by_pred(samples, eval_sample_func):
    all_sample_attentions = average_sample_attention_scores(samples) 
    samples_pred_true = [
        sample for sample in samples if eval_sample_func(sample) 
    ]
    samples_pred_false = [
        sample for sample in samples if not eval_sample_func(sample)
    ]
    true_sample_attentions = average_sample_attention_scores(samples_pred_true)
    false_sample_attentions = average_sample_attention_scores(samples_pred_false)

    for section in list(true_sample_attentions.keys()):
        new_section = f"{section}_by_pred"
        true_sample_attentions[new_section] = true_sample_attentions.pop(section)
        for key in true_sample_attentions[new_section]:
            new_key = f"{key}_true"
            true_sample_attentions[new_section][new_key] = true_sample_attentions[new_section].pop(key)

    for section in list(false_sample_attentions.keys()):
        new_section = f"{section}_by_pred"
        false_sample_attentions[new_section] = false_sample_attentions.pop(section)
        for key in false_sample_attentions[new_section]:
            new_key = f"{key}_false"
            false_sample_attentions[new_section][new_key] = false_sample_attentions[new_section].pop(key)
    
    pred_sample_attentions = {}
    for new_section in true_sample_attentions:
        pred_sample_attentions[new_section] = true_sample_attentions[new_section] | false_sample_attentions[new_section]

    return all_sample_attentions | pred_sample_attentions

def prepare_biasbios_few_shot_examples(
        fewshot_samples, text_key="prompt" , example_sep = "\n###\n", text_target_sep = " "
    ):
    labeled_examples = [] 
    for example in fewshot_samples:
        text = example[text_key]
        entity = example['entity']
        target = example['target_mediated'] 
        text_label = text + text_target_sep + target 
        labeled_examples.append(text_label) 
    
    fewshot_examples = example_sep.join(labeled_examples) + example_sep
    return fewshot_examples 

def prepare_counterfact_fewshot_examples(
        fewshot_samples, 
        add_unmediated_fact="left", unmediated_prefix="Previously ", mediated_prefix="Currently ",
        example_sep = "\n###\n", text_target_sep = " "
    ):
    labeled_examples = [] 
    for example in fewshot_samples:
        prompt = example['prompt']
        context = example['context']
        target_mediated = example['target_mediated']
        target_unmediated = example['target_unmediated']

        if add_unmediated_fact == "left":
            unmediated_fact = context.replace(target_mediated, target_unmediated)+". "
            new_prompt = f"{unmediated_prefix}{unmediated_fact}{mediated_prefix}{prompt}"
        elif add_unmediated_fact == "right":
            prompt = prompt.replace(target_mediated, target_unmediated)
            context = context.rstrip(". ") + ". "
            new_prompt = f"{mediated_prefix}{context}{unmediated_prefix}{prompt}"
        else:
            raise ValueError("Unexpected %s"%(add_unmediated_fact))
        
        text_label = new_prompt + text_target_sep + target_mediated
        labeled_examples.append(text_label) 
    
    fewshot_examples = example_sep.join(labeled_examples) + example_sep
    return fewshot_examples 


        






    
    
        


            

        














"""Utilities to make using python logging module easier."""
import argparse
import logging
import sys
from typing import Any

DEFAULT_FORMAT = "%(asctime)s %(name)s %(levelname)-8s %(message)s"
DEFAULT_DATEFMT = "%Y-%m-%d %H:%M:%S"
DEFAULT_LEVEL = logging.INFO


def configure(args: argparse.Namespace | None = None, **kwargs: Any) -> None:
    """Configure logging globally with reasonable defaults."""
    if "level" in kwargs:
        level = kwargs.get("level", DEFAULT_LEVEL)
    elif args is not None:
        level = getattr(args, "log_level", DEFAULT_LEVEL)
    else:
        level = DEFAULT_LEVEL

    kwargs.setdefault("stream", sys.stdout)
    kwargs.setdefault("format", DEFAULT_FORMAT)
    kwargs.setdefault("datefmt", DEFAULT_DATEFMT)
    kwargs.setdefault("level", level)
    logging.basicConfig(**kwargs)


def add_logging_args(parser: argparse.ArgumentParser) -> None:
    """Add --verbose (-v) and --quiet (-q) args to the parser."""
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_const",
        dest="log_level",
        const=logging.DEBUG,
        default=logging.INFO,
        help="show all log statements",
    )
    parser.add_argument(
        "--quiet",
        "-q",
        action="store_const",
        dest="log_level",
        const=logging.WARNING,
        help="hide all non-critical messages",
    )



"""Tools for reading and controlling the runtime environment."""
import os
import pathlib

from evaluation.utils.typing import PathLike

ENV_DATA_DIR = "CM_DATA_DIR"
ENV_MODELS_DIR = "CM_MODELS_DIR"
ENV_RESULTS_DIR = "CM_RESULTS_DIR"

DEFAULT_DATA_DIR = "data"
DEFAULT_MODELS_DIR = "models"
DEFAULT_RESULTS_DIR = "results"


def maybe_relative_to_repo(path: PathLike) -> pathlib.Path:
    """Resolve the (potentially relative) path.

    Args:
        path (PathLike): The path to resolve. If the path is relative, it is
            assumed to be relative to the repository root. If this is already
            an absolute path, it is returned unchanged.

    Returns:
        pathlib.Path: The resolved path.

    """
    path = pathlib.Path(path)
    if path.is_absolute():
        return path
    return pathlib.Path(__file__).parents[2] / path


def read_path(name: str, default: PathLike) -> pathlib.Path:
    """Try to read a path from the env.

    Args:
        name (str): Name of the env variable to read.
        default (PathLike): The default path in case the environment variable
            does not exist. If relative, assumed to be relative from the
            repo root.

    Returns:
        PathLike: The path, if the variable could be read and/or
            if default was provided.

    """
    read = os.environ.get(name)
    path: PathLike = maybe_relative_to_repo(default) if read is None else read
    return pathlib.Path(path)


def determine_data_dir(default: PathLike = DEFAULT_DATA_DIR) -> pathlib.Path:
    """Return directory containing project datasets.

    Args:
        default (PathLike, optional): Default to use if CM_DATA_DIR env
            variable is not set. Defaults to './data'.

    Returns:
        pathlib.Path: Directory data is stored in.

    """
    return read_path(ENV_DATA_DIR, default)


def determine_models_dir(default: PathLike = DEFAULT_MODELS_DIR) -> pathlib.Path:
    """Return directory containing project models.

    Args:
        default (PathLike, optional): Default to use if CM_MODELS_DIR env
            variable is not set. Defaults to './models'.

    Returns:
        pathlib.Path: Directory data is stored in.

    """
    return read_path(ENV_MODELS_DIR, default)


def determine_results_dir(default: PathLike = DEFAULT_RESULTS_DIR) -> pathlib.Path:
    """Return directory containing results from any scripts.

    Args:
        default (PathLike, optional): Default to use if CM_RESULTS_DIR env
            variable is not set. Defaults to './results'.

    Returns:
        pathlib.Path: Directory results are stored in.

    """
    return read_path(ENV_RESULTS_DIR, default)



"""Evaluate editor effects on generation for bias setting."""
import argparse
import json
import logging
from pathlib import Path

from evaluation import benchmarks, data, models, precompute
from evaluation.utils import experiment_utils, logging_utils
from pastalib import pasta

import torch
from torch.utils.tensorboard import SummaryWriter

logger = logging.getLogger(__name__)


def main(args: argparse.Namespace) -> None:
    """Run the evaluation for instruction following tasks."""
    experiment = experiment_utils.setup_experiment(args)
    logging_utils.configure(args=args)
    data.disable_caching()
    if args.debug: 
        experiment_utils.set_ipdb_trace()

    # Set up the model and tokenzier 
    device = args.device or "cuda" if torch.cuda.is_available() else "cpu"
    mt = models.load_model(args.model, device=device, fp16=args.fp16)

    # Initialize the PASTA steerer 
    if args.apply_pasta:
        head_config = experiment_utils.read_head_config(args.pasta_head_config)
        pasta_steerer = pasta.PASTA(
            mt.model, 
            mt.tokenizer,
            head_config=head_config, 
            alpha=args.alpha, 
            scale_position=args.scale_position,
        )
    else:
        pasta_steerer = None

    # Set up the evaluation data 
    if args.example_subset is not None:
        split = f"train[{args.example_subset}]"
    else:
        split = "train[5000:10000]"
    dataset = data.load_dataset("biosbias", file=args.dataset_file, split=split)
    dataset = precompute.from_args(args, dataset)
    dataset = precompute.prompt_in_context_from_dataset(
        dataset, output_key="prompt", context_suffix="\n\n"
    )
    
    # Benchmark evaluation arguments 
    benchmark_kwargs: dict = {}
    benchmark_kwargs["emphasized_text"] = args.emphasized_text 
    benchmark_kwargs["batch_size"] = args.batch_size 
    benchmark_kwargs["max_new_tokens"] = args.max_new_tokens 
    benchmark_kwargs["entity_occurrence"] = 1
    benchmark_kwargs["add_few_shot"] = args.add_few_shot 
    benchmark_kwargs["few_shot_index"] = args.few_shot_index 
    if args.max_length is not None:
        benchmark_kwargs["max_length"] = args.max_length 
    tfidf_vectorizer = data.load_biosbias_tfidf_vectorizer()
    benchmark_kwargs["tfidf_vectorizer"] = tfidf_vectorizer
    # Set up output dir 
    output_dir = "{pasta}_{update_section}".format(
        pasta = f"pasta-{args.alpha}-{args.scale_position}" if args.apply_pasta else "pasta-none",
        update_section = f"update-{args.emphasized_text}" if args.emphasized_text else "update-none", 
    )
    if args.add_few_shot:
        output_dir = output_dir + "_few-%d-%s"%(args.add_few_shot, str(args.few_shot_index))
    evaluation_result_dir = ( 
        experiment.results_dir 
        / f"{args.task}" 
        / "{model}_prompt-{prompt_idx}_data-{example_subset}".format(
            model = args.model.split("/")[-1], 
            prompt_idx = f"{','.join([str(idx) for idx in args.prompt_idx])}" if args.prompt_idx else "all",
            example_subset = args.example_subset if args.example_subset else "all", 
            )
        / output_dir 
    )
    if args.pasta_head_config is not None:
        evaluation_result_dir = evaluation_result_dir / f"{args.pasta_head_config.split('/')[-1]}"

    evaluation_result_dir.mkdir(exist_ok=True, parents=True) 
    result_file = evaluation_result_dir / f"instruction_evaluation_{args.task}.json"
    metric_file = evaluation_result_dir / "metric_result.json" 

    if not result_file.exists() or args.overwrite_output_dir:
        logger.info("begin baseline")
        evluation_result = benchmarks.biasbios_instruction_evaluation(
            mt=mt,
            dataset=dataset,
            task=args.task, 
            prompt_idx=args.prompt_idx, 
            pasta_steerer=pasta_steerer, 
            device=device,
            desc="Instruction evaluation [LM]",
            **benchmark_kwargs,
        )
        logging.info(
            f"Evaluation complete! results:\n%s",
            json.dumps(evluation_result.metrics.to_dict(), indent=1),
        )
        tb_writter = SummaryWriter(log_dir=str(evaluation_result_dir.resolve()))

        metrics = evluation_result.metrics.to_dict() 
        tb_writter.add_scalar("top1_accuracy", metrics['top1_accuracy'], 1)
        tb_writter.add_scalar(f"top{metrics['k']}_accuracy", metrics['topk_accuracy'], 1)
        for key in ["mean", "std"]:
            tb_writter.add_scalar(f"fluency/{key}", metrics['fluency'][key], 1)
            tb_writter.add_scalar(f"consistency/{key}", metrics['consistency'][key], 1)
        instruction_evaluation_result = metrics['instruction_evaluation'] 
        for key in instruction_evaluation_result:
            tb_writter.add_scalar(f"instruction_evaluation/{key}", instruction_evaluation_result[key], 1)
        tb_writter.close()

        result_file.parent.mkdir(exist_ok=True, parents=True)
        with result_file.open("w") as handle:
            json.dump(evluation_result.to_dict(), handle)
        with metric_file.open("w") as handle:
            json.dump(metrics, handle) 
    else:
        logger.info(
            f"existing baseline results found at {result_file}; skipping"
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="evaluate editor generation on bias dataset"
    )
    parser.add_argument(
        "--task", type=str, default="json", help="The name of evaluation task: [json|pronchange]."
    )
    parser.add_argument(
        "--prompt_idx", nargs="+", type=int, default=0, help="Which prompt template to apply for evaluation."
    )
    parser.add_argument(
        "--dataset_file", type=Path, help="path to dataset.", default=None, 
    )
    models.add_model_args(parser)
    precompute.add_preprocessing_args(parser)
    experiment_utils.add_experiment_args(parser)
    logging_utils.add_logging_args(parser)
    args = parser.parse_args()
    main(args)



"""Reformat datasets.

This script handles reformatting and caching dataset from outside sources.
Callers usually should specify --dataset-file and point to the raw data
downloaded from its source.

Once the data is reformatted, you'll no longer have to specify --dataset-file
to any other scripts. The code will simply read it from the cache.
"""
import argparse
from pathlib import Path

from evaluation import data
from evaluation.utils import logging_utils


def main(args: argparse.Namespace) -> None:
    """Do the reformatting by loading the dataset once."""
    # data.disable_caching()
    logging_utils.configure(args=args)
    # data.load_dataset(args.dataset, file=args.dataset_file)

    file = data._reformat_bias_in_bios_file(
        args.biasbios_raw_path, bio_min_words=10, sent_min_words=3, 
        file_name=args.biasbios_save_file, sents_choice="all", 
        attr_sent_idx=0,
    )
    return file 



if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="reformat datasets on disk")
    data.add_dataset_args(parser)
    parser.add_argument(
        "--biasbios_raw_path", type=Path, help="path to BiasBios raw file.", default=None, 
    )
    parser.add_argument(
        "--biasbios_save_file", type=Path, help="path to save the dataset.", default=None, 
    )
    logging_utils.add_logging_args(parser)
    args = parser.parse_args()
    main(args)



"""Datasets for evaluating context mediation in LMs."""
import argparse
import csv
import json
import logging
import pickle
import random
from collections import defaultdict
from functools import cache
from itertools import chain
from pathlib import Path
from typing import Any, Sequence, TypedDict, cast

from evaluation.utils import env_utils, lang_utils
from evaluation.utils.typing import Dataset, PathLike, StrSequence

import datasets
import numpy
import scipy.sparse
import spacy
import wget
from sklearn.feature_extraction.text import TfidfVectorizer
from tqdm.auto import tqdm

logger = logging.getLogger(__name__)


SUPPORTED_DATASETS = ("counterfact", "winoventi", "biosbias", "mcrae")

ROME_BASE_URL = "https://rome.baulab.info/data/dsets"
COUNTERFACT_URL = f"{ROME_BASE_URL}/counterfact.json"
ATTRIBUTE_SNIPPETS_URL = f"{ROME_BASE_URL}/attribute_snippets.json"
TFIDF_IDF_URL = f"{ROME_BASE_URL}/idf.npy"
TFIDF_VOCAB_URL = f"{ROME_BASE_URL}/tfidf_vocab.json"

WINOVENTI_URL = "https://raw.githubusercontent.com/commonsense-exception/commonsense-exception/main/data/winoventi_bert_large_final.tsv"

_MCRAE_BLACKLISTED_FEATURE_PREFIXES = ("bought/sold", "eg -", "killed", "king of")
_MCRAE_SPLITTABLE_FEATURE_PREFIXES = (
    "associated with",
    "an",
    "a",
    "becomes a",
    "causes",
    "comes from",
    "comes in",
    "comes on",
    "different",
    "found at",
    "found below",
    "found by",
    "found in",
    "found on",
    "found over",
    "found near",
    "has an",
    "has a",
    "has",
    "is an",
    "is attached to",
    "is a",
    "is",
    "like a",
    "made by",
    "made of",
    "made with",
    "made from",
    "owned by",
    "part of a",
    "part of",
    "requires a",
    "requires",
    "used as",
    "used at",
    "used by",
    "used for",
    "used in",
    "used on",
    "used with",
    "uses",
)

_BIOS_BIAS_BLACKLISTED_NAMES = frozenset(
    {
        "Non-Residential",
    }
)

# These prefixes do not make as much sense when put in front of the first name, so
# we'll try to remove them as much as possible.
_BIOS_BIAS_PREFIXES = (
    "professor",
    "prof.",
    "prof",
    "dr.",
    "dr",
    "doctor",
    "mr.",
    "mr",
    "ms.",
    "ms",
    "mrs.",
    "mrs",
    "rev.",
    "rev",
    "pastor",
)

_COUNTERFACT_PARAPHRASE_PROMPT_ARTIFACTS = (" (b. ", "(tr. ", "(min. ")


class ContextMediationSample(TypedDict):
    """Single sample that can be used for context mediation analysis."""

    id: str  # Identifier
    entity: str  # "Barack Obama"
    attribute: str  # "invented the iPhone"
    context: str  # "Everyone knows that Barack Obama invented the iPhone."
    prompt: str  # "Barack Obama received a degree in"
    target_mediated: str | None  # "computer science" or not set for generation
    target_unmediated: str | None  # "law" or not set for generation
    source: dict | None  # Where this sample was derived from, e.g. counterfact sample.


class ContextMediationBatch(TypedDict):
    """Batch of context mediation samples."""

    id: StrSequence
    entity: StrSequence
    attribute: StrSequence
    context: StrSequence
    prompt: StrSequence
    target_mediated: StrSequence | None
    target_unmediated: StrSequence | None
    source: Sequence[dict] | None


ContextMediationInput = ContextMediationSample | ContextMediationBatch


def _determine_file(file: PathLike | None, url: str) -> Path:
    """Set default for (maybe null) file to something sensible based on the URL."""
    if file is None:
        name = url.split("/")[-1]
        file = env_utils.determine_data_dir() / name
    return Path(file).resolve()


def _download_file(file: PathLike, url: str) -> None:
    """Download the url to file."""
    file = Path(file)
    file.parent.mkdir(exist_ok=True, parents=True)
    wget.download(url, out=str(file))


def _rejoin_sents_on_entity(entity: str, sents: list[str]) -> list[str]:
    """Rejoin any splits where the entity was broken into multiple sentences."""
    candidates = [
        index
        for index, (left, right) in enumerate(zip(sents, sents[1:]))
        if (entity not in left and entity not in right and entity in f"{left} {right}")
        or left.endswith(entity)
    ]
    if not candidates:
        return sents

    assert len(candidates) == 1
    [index] = candidates
    merged = f"{sents[index]} {sents[index + 1]}"
    return [*sents[:index], merged, *sents[index + 2 :]]


def _strip_counterfact_paraphrase_prompt(entity: str, prompt: str) -> str:
    """Strip the cruft out of one of CounterFact's paraphrase prompts."""
    nlp = load_spacy_model("en_core_web_sm")

    # Sometimes the paraphrase model stuck in a bunch of newlines, always
    # take what comes after them.
    prompt = prompt.split("\n")[-1].strip()

    # Similarly, it sometimes adds a wikipedia title. In these cases, a good
    # heuristic is just to only take everything beyond and including the entity.
    if prompt.startswith("Category:"):
        prompt = entity + prompt.split(entity)[-1]

    # Another common, weird artifact to get rid of:
    for artifact in _COUNTERFACT_PARAPHRASE_PROMPT_ARTIFACTS:
        if artifact in prompt:
            prompt = prompt.split(artifact)[-1]

    sents = [str(sent) for sent in nlp(prompt).sents]
    sents = _rejoin_sents_on_entity(entity, sents)
    if len(sents) <= 2:
        if entity not in sents[0]:
            assert len(sents) == 2
            sents = [sents[1]]
    else:
        if "?" in sents[-2]:
            sents = [sents[-2], sents[-1]]
        else:
            sents = [sents[-1]]

    stripped = " ".join(sents).strip()

    # If there's still a period in the sentence, but the entity does not have it,
    # it usually means we failed to split a sentence.
    if "." in stripped and "." not in entity:
        stripped = stripped.split(".")[-1].strip()

    # Finally, if we really messed something up along the way, just default to the
    # cleaned up paraphrase prompt.
    if entity not in stripped:
        logger.debug(f"prompt cleaning failed for {entity}: {stripped}")
        return prompt

    return stripped


def _reformat_counterfact_sample(cf_sample: dict) -> ContextMediationSample:
    """Reformat the counterfact sample."""
    cf_case_id = cf_sample["case_id"]
    cf_requested_rewrite = cf_sample["requested_rewrite"]
    cf_subject = cf_requested_rewrite["subject"]
    cf_target_new = cf_requested_rewrite["target_new"]["str"]
    cf_target_true = cf_requested_rewrite["target_true"]["str"]
    cf_prompt = cf_requested_rewrite["prompt"].format(cf_subject)
    cf_paraphrase_prompts = cf_sample["paraphrase_prompts"]

    entity = cf_subject
    prompt = _strip_counterfact_paraphrase_prompt(entity, cf_paraphrase_prompts[0])
    context = f"{cf_prompt} {cf_target_new}"
    attribute = context.split(entity)[-1].strip(",-;: ")
    target_mediated = cf_target_new
    target_unmediated = cf_target_true

    return ContextMediationSample(
        id=str(cf_case_id),
        entity=entity,
        prompt=prompt,
        context=context,
        attribute=attribute,
        target_mediated=target_mediated,
        target_unmediated=target_unmediated,
        # NOTE(evandez): Need to copy or else remove_columns will directly
        # delete keys on the original dict, causing source to be empty dict.
        source={**cf_sample},
    )


def _reformat_counterfact_file(file: Path) -> Path:
    """Reformat the counterfact file to be jsonl instead of json."""
    with file.open("r") as handle:
        lines = json.load(handle)
    file = file.parent / f"{file.stem}.jsonl"
    with file.open("w") as handle:
        for line in tqdm(lines, desc="reformat counterfact"):
            json.dump(_reformat_counterfact_sample(line), handle)
    return file


def _load_counterfact(
    file: PathLike | None = None,
    url: str = COUNTERFACT_URL,
    overwrite: bool = False,
    **kwargs: Any,
) -> Dataset:
    """Download and format the counterfact dataset."""
    if file is None:
        file = env_utils.determine_data_dir() / "counterfact.jsonl"

        # If jsonl file not here, means we need to download and reformat.
        if not file.exists():
            file = _determine_file(None, url)
            if not file.exists() or overwrite:
                _download_file(file, url)

    file = Path(file)
    if file.suffix != ".jsonl":
        file = _reformat_counterfact_file(file)

    dataset = datasets.load_dataset("json", data_files=str(file), **kwargs)
    assert isinstance(
        dataset, datasets.arrow_dataset.Dataset | datasets.dataset_dict.DatasetDict
    ), type(dataset)

    return dataset


def _filter_winoventi_sample(wv_sample: dict) -> bool:
    """Determine whether the WV sample is well-formed; some have typos."""
    wv_word = wv_sample["Word"]
    wv_masked_prompt = wv_sample["masked_prompt"]
    wv_biased_context = wv_sample["biased_word_context"]
    wv_adversarial_context = wv_sample["adversarial_word_context"]

    prompt_components = wv_masked_prompt.split(". ")
    has_no_extra_period = len(prompt_components) == 2
    has_two_entity_mentions = wv_masked_prompt.count(wv_word) == 2

    has_entity_before_attribute = False
    if has_no_extra_period:
        context, _ = prompt_components
        if wv_word in context:
            start = context.index(wv_word)
            has_entity_before_attribute = wv_biased_context in context[start:]
            has_entity_before_attribute |= wv_adversarial_context in context[start:]

    return (
        has_no_extra_period and has_two_entity_mentions and has_entity_before_attribute
    )


def _reformat_winoventi_sample(wv_sample: dict) -> ContextMediationSample:
    """Reformat winoventi sample to use context mediation format."""
    wv_word = entity = wv_sample["Word"]

    wv_masked_prompt = wv_sample["masked_prompt"]
    wv_prompt_components = wv_masked_prompt.split(". ")
    assert len(wv_prompt_components) == 2, wv_masked_prompt

    context, prompt = wv_prompt_components
    assert entity in context, context
    assert entity in prompt, prompt
    prompt = prompt.replace("[MASK]", "").strip(". ")
    context_components = context.split(entity)
    assert len(context_components) == 2
    attribute = context_components[1].strip(". ")

    wv_target = wv_sample["target"]
    wv_incorrect = wv_sample["incorrect"]

    wv_id = f"{wv_word}_{wv_target}_{wv_incorrect}"

    return ContextMediationSample(
        id=wv_id,
        entity=entity,
        prompt=prompt,
        context=context,
        attribute=attribute,
        target_mediated=wv_target,
        target_unmediated=wv_incorrect,
        source={**wv_sample},
    )


def _load_winoventi(
    file: PathLike | None = None,
    url: str = WINOVENTI_URL,
    overwrite: bool = False,
    **kwargs: Any,
) -> Dataset:
    """Download and reformat the winoventi dataset."""
    file = _determine_file(file, url)
    if not file.exists() or overwrite:
        _download_file(file, url)

    dataset = datasets.load_dataset(
        "csv", data_files=str(file), delimiter="\t", **kwargs
    )
    assert isinstance(
        dataset, datasets.arrow_dataset.Dataset | datasets.dataset_dict.DatasetDict
    ), type(dataset)

    dataset = dataset.filter(_filter_winoventi_sample, desc="filter winoventi")
    dataset = dataset.map(
        _reformat_winoventi_sample,
        remove_columns=column_names(dataset),
        desc="reformat winoventi",
    )

    return dataset


def _get_attribute(
    bb_bio:str, 
    bb_name:str,
    nlp, 
    sent_idx: int|None=None, 
):
    if sent_idx is None:
        attribute = bb_bio[bb_bio.index(bb_name) + len(bb_name) :].strip(".: ")
    else:
        doc = nlp(bb_bio)
        sents = [sent for sent in doc.sents]
        attr_sent =  str(sents[sent_idx])
        if sent_idx == 0 and bb_name in attr_sent:
            attribute = attr_sent[attr_sent.index(bb_name) + len(bb_name) :].strip(".: ")
        else:
            attribute = attr_sent
    return attribute


def _reformat_bias_in_bios_file(
    pkl_file: Path,
    bio_min_words: int = 10,
    sent_min_words: int = 3,
    limit: int | None = 50000,
    file_name: str = "biosbias.json",
    sents_choice: int | str = 1, 
    attr_sent_idx: int | None = None, 
) -> Path:
    """Reformat the Bias in Bios pickle file on disk."""
    with pkl_file.open("rb") as handle:
        data = pickle.load(handle)

    if limit is not None:
        data = data[:limit]

    # We'll use first name as the entity.
    bb_names = [sample["name"][0] for sample in data]

    # Take only one sentence of each bio to make the task harder.
    nlp = load_spacy_model("en_core_web_sm")
    bb_bios_raw = [sample["raw"] for sample in data]

    bb_bios_abridged = []
    for doc in tqdm(nlp.pipe(bb_bios_raw), total=len(data), desc="parse biosbias"):
        sents = [
            str(sent)
            for sent in doc.sents
            if len(str(sent).strip().split()) >= sent_min_words
        ]
        if len(sents) == 0:
            bb_bio_abridged = ""  # will get filtered out later
        elif len(sents) == 1:
            bb_bio_abridged = sents[0]  # no choice but first
        else:
            if sents_choice == "all":
                bb_bio_abridged = " ".join(sents)
            else:
                bb_bio_abridged = sents[sents_choice]  # otherwise, always take second sentence
        bb_bios_abridged.append(bb_bio_abridged)

    # Normalize the samples.
    lines = []
    for index, (sample, bb_name, bb_bio) in enumerate(
        tqdm(zip(data, bb_names, bb_bios_abridged), total=len(data), desc="process")
    ):
        if bb_name in _BIOS_BIAS_BLACKLISTED_NAMES:
            logger.debug(
                f"will not include sample #{index} because it has "
                f"blacklisted name '{bb_name}'"
            )
            continue

        # Replace all variants of the person's name with their first name.
        # We'll ignore suffixes, but try to strip prefixes since they are much more
        # common and do not mesh well with just first names.
        full_name = " ".join(part for part in sample["name"] if part)
        last_name = sample["name"][-1]
        first_last_name = f"{bb_name} {last_name}"

        replacements = []
        for name in (full_name, first_last_name, last_name):
            for prefix in _BIOS_BIAS_PREFIXES:
                replacements.append(f"{prefix} {name}")
                replacements.append(f"{prefix.capitalize()} {name}")
                replacements.append(f"{prefix.upper()} {name}")
        replacements += [full_name, first_last_name, last_name]
        for replace in replacements:
            bb_bio = bb_bio.replace(replace, bb_name)
        bb_bio = bb_bio.strip("*• ")

        bb_title = sample["title"].replace("_", " ")
        bb_id = "_".join(part for part in sample["name"] if part)

        # NOTE(evan): This filtering is necessary due to a downstream limitation where
        # the editor assumes that the last occurrence of an entity is always the second
        # occurrence.
        n_occurrences = bb_bio.count(bb_name)
        if n_occurrences == 0:
            bb_bio = f"About {bb_name}: {bb_bio}"
        elif n_occurrences != 1:
            logger.debug(
                f"will not include sample #{index} because there are "
                f"{n_occurrences} (!= 1) occurrences of '{bb_name}' in '{bb_bio}'"
            )
            continue

        approx_n_words = len(bb_bio.split())
        if approx_n_words < bio_min_words:
            logger.debug(
                f"will not include sample #{index} because bio '{bb_bio}' contains "
                f"too few words (approx. < {bio_min_words})"
            )
            continue

        entity = bb_name
        prompt = f"{entity} has the occupation of"
        context = bb_bio
        # attribute = bb_bio[bb_bio.index(bb_name) + len(bb_name) :].strip(".: ")
        attribute = _get_attribute(bb_bio, bb_name, nlp, sent_idx=attr_sent_idx)
        target_mediated = bb_title

        line = ContextMediationSample(
            id=bb_id,
            source=sample,
            entity=entity,
            prompt=prompt,
            context=context,
            attribute=attribute,
            target_mediated=target_mediated,
            target_unmediated=None,
        )
        lines.append(line)

    # Save in jsonl format.
    json_file = env_utils.determine_data_dir() / file_name
    with json_file.open("w") as handle:
        for line in lines:
            json.dump(dict(line), handle)
    return json_file


def _load_bias_in_bios(file: PathLike | None = None, **kwargs: Any) -> Dataset:
    """Load the Bias in Bios datast, if possible."""
    if file is None:
        logger.debug("biosbias file not set; defaulting to environment data dir")
        file = env_utils.determine_data_dir() / "biosbias.json"

    file = Path(file)
    if not file.exists():
        raise FileNotFoundError(
            f"biosbias file not found: {file}"
            "\nnote this dataset cannot be downloaded automatically, "
            "so you will have to retrieve it by following the instructions "
            "at the github repo https://github.com/microsoft/biosbias and pass "
            "the .pkl file in via the -f flag"
        )

    if file.suffix in {".pkl", ".pickle"}:
        logger.debug(f"{file} is pickle format; will reformat into json")
        file = _reformat_bias_in_bios_file(file)

    dataset = datasets.load_dataset("json", data_files=str(file), **kwargs)
    assert isinstance(
        dataset, datasets.arrow_dataset.Dataset | datasets.dataset_dict.DatasetDict
    ), type(dataset)
    return dataset


def _get_mcrae_concept(row: dict) -> str:
    """Return the McRae concept, cleaning it up."""
    concept = row["Concept"]
    return concept.replace("_", " ")


def _get_mcrae_feature(row: dict) -> str:
    """Return the normalized McRae feature, cleaning it up."""
    feature = row["Feature"]
    for prefix in ("inbeh", "beh"):
        feature = feature.replace(f"{prefix}_-_", "")
    feature = feature.replace("_", " ")
    return feature


def _get_mcrae_feature_prob(row: dict) -> float:
    """Get p(feature | concept) using production frequencies."""
    return int(row["Prod_Freq"]) / 30


def _get_mcrae_sample_id(
    concept: str, context_feature: str, prompt_feature: str
) -> str:
    """Return a unique, readable ID for the McRae context mediation sample."""
    cid = concept.replace(" ", "_")
    cfid = context_feature.replace(" ", "_")
    pfid = prompt_feature.replace(" ", "_")
    return f"{cid}/{cfid}/{pfid}"


def _filter_mcrae_features(rows: list[dict]) -> list[dict]:
    """Filter out any rows with blacklisted features."""
    filtered = []
    for row in rows:
        feature = _get_mcrae_feature(row)
        if not any(map(feature.startswith, _MCRAE_BLACKLISTED_FEATURE_PREFIXES)):
            filtered.append(row)
    return filtered


def _get_mcrae_feature_prefix_for_fluency(feature: str) -> str | None:
    """Return True if feature needs to be prefixed with "is" to be fluent."""
    if any(
        feature.startswith(prefix)
        for prefix in {
            "a ",
            "an ",
            "associated ",
            "hunted ",
            "eaten ",
            "found ",
            "like ",
            "made ",
            "part ",
            "owned ",
            "used ",
            "worn ",
        }
    ):
        return "is"
    elif any(feature.startswith(prefix) for prefix in ("different ",)):
        return "has"
    return None


def _make_mcrae_feature_fluent(feature: str) -> str:
    """Heuristically adjust feature so that concept can be fluently prepended."""
    prefix = _get_mcrae_feature_prefix_for_fluency(feature)
    if prefix is not None:
        feature = f"{prefix} {feature}"
    return feature


def _strip_mcrae_parenthetical(concept: str) -> str:
    """Strip parenthetical from concept."""
    return concept.split("(")[0].strip()


def _get_mcrae_prompt_and_target(feature: str) -> tuple[str, str]:
    """Transform McRae concept and target feature to a prompt and target token."""
    # We will split the feature into a prompt part (what is fed into the LM)
    # and a target part (the probability of which is maximized via REMEDI and
    # measured during eval). This split depends on how the feature is phrased,
    # as we want to avoid cases where the prompt inadvertently pigeonholes the LM
    # into saying the correct thing.
    components = None
    for prefix in _MCRAE_SPLITTABLE_FEATURE_PREFIXES:
        prefix = f"{prefix.strip()} "
        if feature.startswith(prefix):
            logger.debug(f'splitting feature "{feature}" on "{prefix}"')
            components = [
                _make_mcrae_feature_fluent(prefix).lstrip(),
                feature[len(prefix) :].strip(),
            ]
            break

    if components is None:
        logger.debug(f'feature "{feature}" will be the full target')
        prompt = _get_mcrae_feature_prefix_for_fluency(feature)
        if prompt is None:
            prompt = ""
        components = [prompt, feature]

    [prompt, target] = components
    return prompt, target


def _get_mcrae_prompt_with_entity(concept: str, prompt: str) -> str:
    """Glue the concept and prompt together."""
    a = lang_utils.determine_article(concept).capitalize()
    prompt = f"{a} {concept} {prompt}".strip()
    return prompt


def _get_mcrae_context_and_attribute(concept: str, feature: str) -> tuple[str, str]:
    """Transform McRae concept and feature to a context and attribute."""
    feature = _make_mcrae_feature_fluent(feature)
    a = lang_utils.determine_article(concept).capitalize()
    context = f"{a} {concept} {feature}."
    return context, feature


def _create_samples_from_mcrae_norms(
    text_file: Path,
    min_co_prob: float = 0.1,
    samples_per_feature_pair: int = 1,
    unrelated_features_per_sample: int = 5,
    seed: int | None = 123456,
) -> Path:
    """Transform McRae entries into context mediation samples.

    The McRae norms dataset gives us concepts and features of those concepts, e.g.,
    a concept might be "accordion" and a feature for it would be "musical instrument".

    To create context mediation samples, we look for pairs of attributes that often
    co-occur for different concepts. We make one the context and one the prompt,
    selecting the concept to apply it to at random from the set of concepts that do
    not have either attribute. In the "source" field, we store all features implied
    by the context as well as the probability that they co-occur.
    """
    with text_file.open("r") as handle:
        rows = list(csv.DictReader(handle, delimiter="\t"))
    rows = _filter_mcrae_features(rows)

    # Index features and concepts.
    features_by_concept = defaultdict(set)
    concepts_by_feature = defaultdict(set)
    probs_by_concept_feature = defaultdict(float)
    prompts_by_feature = {}
    targets_by_feature = {}
    for row in rows:
        concept = _get_mcrae_concept(row)
        feature = _get_mcrae_feature(row)
        prob = _get_mcrae_feature_prob(row)
        prompt, target = _get_mcrae_prompt_and_target(feature)

        concepts_by_feature[feature].add(concept)
        features_by_concept[concept].add(feature)
        probs_by_concept_feature[concept, feature] = prob
        prompts_by_feature[feature] = prompt
        targets_by_feature[feature] = target

    # Find attributes that co-occur.
    f_o: dict[str, int] = defaultdict(int)
    f_co: dict[str, dict[str, float]] = defaultdict(lambda: defaultdict(float))
    for row in rows:
        concept = _get_mcrae_concept(row)
        feature = _get_mcrae_feature(row)
        prob = _get_mcrae_feature_prob(row)

        others = features_by_concept[concept] - {feature}
        for other in others:
            f_co[feature][other] += prob * probs_by_concept_feature[concept, other]
        f_o[feature] += 1

    f_prob = {
        feature: {
            other: f_co[feature][other] / f_o[feature]
            for other in concepts_by_feature
            if other != feature and f_co[feature][other] > 0
        }
        for feature in concepts_by_feature
    }

    # Choose the pairs we're going to use.
    feature_pairs = [
        (feature, other)
        for feature, cos in f_prob.items()
        for other, co_prob in cos.items()
        if co_prob >= min_co_prob
    ]

    # Make the samples. If requested, set seed ahead of time so we always
    # generate the same dataset.
    if seed is not None:
        random.seed(seed)
    random.shuffle(feature_pairs)

    samples = []
    for context_feature, prompt_feature in feature_pairs:
        candidate_concepts = [
            concept
            for concept, features in features_by_concept.items()
            if context_feature not in features and prompt_feature not in features
        ]
        concepts = random.sample(candidate_concepts, k=samples_per_feature_pair)
        for concept in concepts:
            sid = _get_mcrae_sample_id(concept, context_feature, prompt_feature)
            entity = _strip_mcrae_parenthetical(concept)
            prompt = _get_mcrae_prompt_with_entity(
                entity, prompts_by_feature[prompt_feature]
            )
            target_mediated = targets_by_feature[prompt_feature]
            context, attribute = _get_mcrae_context_and_attribute(
                entity, context_feature
            )

            # Sample some features unrelated to the context, prompt, and concept.
            unrelated_features = [
                feature
                for feature in concepts_by_feature
                if feature not in f_prob[context_feature]
                and feature not in features_by_concept[concept]
            ]
            unrelated_features = random.sample(
                unrelated_features, k=unrelated_features_per_sample
            )

            sample = ContextMediationSample(
                id=sid,
                entity=entity,
                context=context,
                attribute=attribute,
                prompt=prompt,
                target_mediated=target_mediated,
                target_unmediated=None,
                source={
                    "concept": concept,
                    "context_feature": context_feature,
                    "context_feature_fluent": _make_mcrae_feature_fluent(
                        context_feature
                    ),
                    "prompt_feature": prompt_feature,
                    "prompt_feature_fluent": _make_mcrae_feature_fluent(prompt_feature),
                    "co_prob": f_prob[context_feature][prompt_feature],
                    "all_co_features": [
                        {
                            "feature": co_feat,
                            "feature_fluent": _make_mcrae_feature_fluent(co_feat),
                            "prompt": _get_mcrae_prompt_with_entity(
                                entity, prompts_by_feature[co_feat]
                            ),
                            "target": targets_by_feature[co_feat],
                            "co_prob": f"{co_prob:.2f}",
                        }
                        for co_feat, co_prob in f_prob[context_feature].items()
                    ],
                    "original_features": [
                        {
                            "feature": feat,
                            "feature_fluent": _make_mcrae_feature_fluent(feat),
                            "prompt": _get_mcrae_prompt_with_entity(
                                entity, prompts_by_feature[feat]
                            ),
                            "target": targets_by_feature[feat],
                            "prob": f"{probs_by_concept_feature[concept, feat]:.2f}",
                        }
                        for feat in features_by_concept[concept]
                    ],
                    "unrelated_features": [
                        {
                            "feature": feat,
                            "feature_fluent": _make_mcrae_feature_fluent(feat),
                            "prompt": _get_mcrae_prompt_with_entity(
                                entity, prompts_by_feature[feat]
                            ),
                            "target": targets_by_feature[feat],
                            # No probability for unrelated features.
                        }
                        for feat in unrelated_features
                    ],
                },
            )
            samples.append(sample)

    json_file = env_utils.determine_data_dir() / "mcrae.json"
    with json_file.open("w") as handle:
        for sample in samples:
            json.dump(sample, handle)

    return json_file


def _load_mcrae(file: PathLike | None = None, **kwargs: Any) -> Dataset:
    """Load the McRae norms dataset, if possible."""
    if file is None:
        logger.debug("mcrae file not set; defaulting to environment data dir")
        file = env_utils.determine_data_dir() / "mcrae.json"

    file = Path(file)
    if not file.exists():
        raise FileNotFoundError(
            f"mcrae file not found: {file}"
            "\nnote this dataset cannot be downloaded automatically, "
            "so you will have to retrieve it from the supplementary material "
            "available here https://link.springer.com/article/10.3758/BF03192726 "
            "and then pass the CONCS_FEATS_concstats_brm.txt file in via the -f flag"
        )

    if file.suffix == ".txt":
        logger.debug(f"{file} is in txt format; will reformat into json")
        file = _create_samples_from_mcrae_norms(file)

    dataset = datasets.load_dataset("json", data_files=str(file), **kwargs)
    assert isinstance(
        dataset, datasets.arrow_dataset.Dataset | datasets.dataset_dict.DatasetDict
    ), type(dataset)
    return dataset


def load_dataset(name: str, **kwargs: Any) -> Dataset:
    """Load the dataset by name."""
    if name == "counterfact":
        return _load_counterfact(**kwargs)
    elif name == "winoventi":
        return _load_winoventi(**kwargs)
    elif name == "biosbias":
        return _load_bias_in_bios(**kwargs)
    elif name == "mcrae":
        return _load_mcrae(**kwargs)
    else:
        raise ValueError(f"unknown dataset: {name}")


AttributeSnippets = dict[str, dict[str, list[dict]]]


def load_attribute_snippets(
    file: Path | None = None, url: str = ATTRIBUTE_SNIPPETS_URL, overwrite: bool = False
) -> AttributeSnippets:
    """Load attribute snippets for different Wikipedia relations/entities.

    This dataset is taken directly from the ROME evaluation. It is not loaded from
    `load_dataset` because it is a mapping, not a sequence. Specifically, it is a
    mapping from Wikidata relation IDs and entity IDs to Wikipedia articles about
    those entities, where the article includes text relevant to the relation. This is
    used to measure consistency of generations with other plausible facts about an
    entity satisfying some relation.

    Args:
        file: Look for attribute snippets at this file. Downloaded otherwise.
        url: Download snippets from this URL.
        overwrite: Overwrite an existing download.

    Returns:
        Mapping from relation ID and entity ID to Wikipedia text.

    """
    file = _determine_file(file, url)
    if not file.exists() or overwrite:
        _download_file(file, url)
    with file.open("r") as handle:
        snippets_list = json.load(handle)

    attribute_snippets: AttributeSnippets = defaultdict(lambda: defaultdict(list))
    for snippets in snippets_list:
        relation_id = snippets["relation_id"]
        target_id = snippets["target_id"]
        for sample in snippets["samples"]:
            attribute_snippets[relation_id][target_id].append(sample)

    return attribute_snippets


def load_counterfact_tfidf_vectorizer(
    idf_file: Path | None = None,
    vocab_file: Path | None = None,
    idf_url: str = TFIDF_IDF_URL,
    vocab_url: str = TFIDF_VOCAB_URL,
    overwrite: bool = False,
) -> TfidfVectorizer:
    """Load precomputed TF-IDF statistics."""
    idf_file = _determine_file(idf_file, idf_url)
    vocab_file = _determine_file(vocab_file, vocab_url)
    for file, url in ((idf_file, idf_url), (vocab_file, vocab_url)):
        if not file.exists() or overwrite:
            _download_file(file, url)

    idf = numpy.load(str(idf_file))
    with vocab_file.open("r") as handle:
        vocab = json.load(handle)

    # Hack borrowed from ROME:
    # https://github.com/kmeng01/rome/blob/0874014cd9837e4365f3e6f3c71400ef11509e04/dsets/tfidf_stats.py#L17
    class ModifiedTfidfVectorizer(TfidfVectorizer):
        TfidfVectorizer.idf_ = idf

    vec = ModifiedTfidfVectorizer()
    vec.vocabulary_ = vocab
    vec._tfidf._idf_diag = scipy.sparse.spdiags(idf, diags=0, m=len(idf), n=len(idf))
    return vec


def load_biosbias_tfidf_vectorizer(
    dataset: datasets.arrow_dataset.Dataset | None = None,
) -> TfidfVectorizer:
    """Load the tfidf vectorizer for Bias in Bios."""
    if dataset is None:
        logger.info("loading full biosbias dataset for tfidf vectorizer")
        dataset = cast(
            datasets.arrow_dataset.Dataset, load_dataset("biosbias", split="train")
        )

    texts = [x["source"]["bio"] for x in dataset]
    logger.info(f"create biosbias tfidf vectorizer from {len(texts)} bios")

    tfidf_vectorizer = TfidfVectorizer()
    tfidf_vectorizer.fit(texts)
    return tfidf_vectorizer


def column_names(dataset: Dataset, exclude: StrSequence | None = None) -> list[str]:
    """Get all column names for the dataset."""
    if isinstance(dataset, datasets.arrow_dataset.Dataset):
        column_names = dataset.column_names
    else:
        assert isinstance(dataset, datasets.dataset_dict.DatasetDict), type(dataset)
        column_names = list(set(chain(*dataset.column_names.values())))

    if exclude is not None:
        column_names = list(set(column_names) - set(exclude))

    return column_names


@cache
def load_spacy_model(name: str) -> spacy.language.Language:
    """Load (and cache) a spacy model."""
    return spacy.load(name)


def maybe_train_test_split(
    dataset: Dataset, **kwargs: Any
) -> datasets.dataset_dict.DatasetDict:
    """Split the dataset into train/test if necessary."""
    if not isinstance(dataset, datasets.dataset_dict.DatasetDict):
        return dataset.train_test_split(**kwargs)
    elif all(key in dataset for key in ("train", "test")):
        return dataset
    elif "train" not in dataset.keys():
        raise ValueError("dataset has no train split?")
    return dataset["train"].train_test_split(**kwargs)


def disable_caching() -> None:
    """Disable all implicit dataset caching."""
    datasets.disable_caching()


def add_dataset_args(
    parser: argparse.ArgumentParser, default: str = "counterfact"
) -> None:
    """Add --dataset and --dataset-file args to the parser."""
    assert default in SUPPORTED_DATASETS, default
    parser.add_argument(
        "--dataset",
        "-d",
        choices=SUPPORTED_DATASETS,
        default=default,
        help="dataset to use",
    )
    parser.add_argument("--dataset-file", "-f", type=Path, help="dataset file to use")



"""Utilities for managing experiment runtimes and results."""
import argparse
import json
import logging
import random
import shutil
import time
from dataclasses import dataclass
from pathlib import Path

from evaluation.utils import env_utils
from evaluation.utils.typing import PathLike

import numpy
import torch

logger = logging.getLogger(__name__)

DEFAULT_SEED = 123456


@dataclass(frozen=True)
class Experiment:
    """A configured experiment."""

    name: str
    results_dir: Path
    seed: int


def read_head_config(pasta_head_config):
    if '.json' in pasta_head_config:
        with open(pasta_head_config, 'r') as handle:
            head_config = json.load(handle) 
    elif '{' in pasta_head_config and '}' in pasta_head_config:
        head_config = json.loads(pasta_head_config)
    else:
        # head_config = [int(h) for h in pasta_head_config.strip().split(',')]
        raise ValueError("Incorrect format of head config.")
    return head_config


def set_seed(seed: int) -> None:
    """Globally set random seed."""
    logger.info("setting all seeds to %d", seed)
    random.seed(seed)
    numpy.random.seed(seed)
    torch.manual_seed(seed)


def set_ipdb_trace():
    import ipdb 
    ipdb.set_trace()


def create_results_dir(
    experiment_name: str,
    root: PathLike | None = None,
    args: argparse.Namespace | None = None,
    args_file_name: str | None = None,
    clear_if_exists: bool = False,
) -> Path:
    """Create a directory for storing experiment results.

    Args:
        name: Experiment name.
        root: Root directory to store results in. Consults env if not set.
        args: If set, save the full argparse namespace as JSON.
        args_file: Save args file here.
        clear_if_exists: Clear the results dir if it already exists.

    Returns:
        The initialized results directory.

    """
    if root is None:
        root = env_utils.determine_results_dir()
    root = Path(root)

    results_dir = root / experiment_name
    results_dir = results_dir.resolve()

    if results_dir.exists():
        logger.info(f"rerunning experiment {experiment_name}")
        if clear_if_exists:
            logger.info(f"clearing previous results from {results_dir}")
            shutil.rmtree(results_dir)

    results_dir.mkdir(exist_ok=True, parents=True)
    if args is not None:
        if args_file_name is None:
            timestamp = time.strftime("%Y%m%d-%H%M%S")
            args_file_name = f"args-{timestamp}.json"
        args_file = results_dir / args_file_name
        logger.info(f"saving args to {args_file}")
        with args_file.open("w") as handle:
            json.dump({key: str(value) for key, value in vars(args).items()}, handle)

    return results_dir


def add_experiment_args(parser: argparse.ArgumentParser) -> None:
    """Add args common to all experiments.

    The args include:
        --experiment-name (-n): Requied, unique identifier for this experiment.
        --results-dir: Root directory containing all experiment folders.
        --clear-results-dir: If set, experiment-specific results directory is cleared.
        --args-file-name: Dump all args to this file; defaults to generated name.
        --seed: Random seed.

    """
    parser.add_argument(
        "--experiment_name",
        "-n",
        required=True,
        help="unique name for the experiment",
    )
    parser.add_argument(
        "--apply_pasta", 
        action="store_true",
        help="Whether to apply PASTA."
    )
    parser.add_argument(
        "--pasta_head_config", 
        type=str, default=None, 
        help="PASTA head config for steering.",
    )
    parser.add_argument(
        "--emphasized_text", 
        type=str, default=None, 
        help="Which textual segments to emphasize:[include|exclude]."
    )
    parser.add_argument(
        "--alpha", 
        type=float, default=None, 
        help="Scaling coefficient."
    )
    parser.add_argument(
        "--scale_position", 
        type=str, default=None, 
        help="Steer the selected section or others."
    )
    parser.add_argument("--add_unmediated_fact", type=bool, default=True, help="Present models both facts.")
    # parser.add_argument("--record_attn_score", nargs="+", type=str, default=None, help="")
    parser.add_argument("--batch_size", type=int, default=16, help="Batch size.")
    parser.add_argument("--max_length", type=int, default=None, help="Max sequence length.")
    parser.add_argument("--max_new_tokens", type=int, default=None, help="Max generation length.")
    parser.add_argument("--add_few_shot", type=int, default=None, help="Apply few-shot prompting.")
    parser.add_argument(
        "--few_shot_index", type=str, default=None, help="Sample index for few shots, e.g., 0,1,2."
    )
    parser.add_argument("--add_marker", type=str, default=None, help="Apply marked prompting.")

    parser.add_argument("--args-file-name", help="file name for args dump")
    parser.add_argument(
        "--example_subset", type=str, default=None, help="run on a subset of data"
    )
    parser.add_argument(
        "--results-dir", type=Path, help="root directory containing experiment results"
    )
    parser.add_argument(
        "--clear-results-dir",
        action="store_true",
        default=False,
        help="clear any old results and start anew",
    )
    parser.add_argument("--overwrite_output_dir", action="store_true", help="")
    parser.add_argument("--seed", type=int, default=DEFAULT_SEED, help="random seed")
    parser.add_argument("--debug", action="store_true", help="Whether to invoke ipdb debugging.")


def setup_experiment(args: argparse.Namespace) -> Experiment:
    """Configure experiment from the args."""
    experiment_name = args.experiment_name
    seed = args.seed

    logger.info(f"setting up experiment {experiment_name}")

    set_seed(seed)

    results_dir = create_results_dir(
        experiment_name,
        root=args.results_dir,
        args=args,
        args_file_name=args.args_file_name,
        clear_if_exists=args.clear_results_dir,
    )

    return Experiment(
        name=experiment_name,
        results_dir=results_dir,
        seed=seed,
    )



"""Utils for interacting with huggingface tokenizers."""
from contextlib import contextmanager
from typing import Any, Iterator, Optional, Sequence, Tuple

from .typing import StrSequence, Tokenizer, TokenizerOffsetMapping


def find_token_range(
    string: str,
    substring: str,
    tokenizer: Optional[Tokenizer] = None,
    occurrence: int = 0,
    offset_mapping: Optional[TokenizerOffsetMapping] = None,
    **kwargs: Any,
) -> Tuple[int, int]:
    """Find index range of tokenized string containing tokens for substring.

    The kwargs are forwarded to the tokenizer.

    A simple example:

        string = 'The batman is the night.'
        substring = 'batman'
        tokenizer = ...

        # Example tokenization: ['the', 'bat', '##man', 'is', 'the', 'night']
        assert find_token_range(string, substring, tokenizer) == (1, 3)

    Args:
        string: The string.
        substring: The substring to find token range for.
        tokenizer: The tokenizer. If not set, offset_mapping must be.
        occurrence: The occurence of the substring to look for.
            Zero indexed. Defaults to 0, the first occurrence.
        offset_mapping: Precomputed offset mapping. If not set, tokenizer will be run.

    Raises:
        ValueError: If substring is not actually in string or if banned
            kwargs are specified.

    Returns:
        Tuple[int, int]: The start (inclusive) and end (exclusive) token idx.
    """
    if tokenizer is None and offset_mapping is None:
        raise ValueError("must set either tokenizer= or offset_mapping=")
    if "return_offsets_mapping" in kwargs:
        raise ValueError("cannot set return_offsets_mapping")
    if substring not in string:
        raise ValueError(f'"{substring}" not found in "{string}"')
    char_start = string.index(substring)
    for _ in range(occurrence):
        try:
            char_start = string.index(substring, char_start + 1)
        except ValueError as error:
            raise ValueError(
                f"could not find {occurrence + 1} occurrences "
                f'of "{substring} in "{string}"'
            ) from error
    char_end = char_start + len(substring)

    if offset_mapping is None:
        assert tokenizer is not None
        tokens = tokenizer(string, return_offsets_mapping=True, **kwargs)
        offset_mapping = tokens.offset_mapping

    token_start, token_end = None, None
    for index, (token_char_start, token_char_end) in enumerate(offset_mapping):
        if token_start is None:
            if token_char_start <= char_start and token_char_end >= char_start:
                token_start = index
        if token_end is None:
            if token_char_start <= char_end and token_char_end >= char_end:
                token_end = index
                break

    assert token_start is not None
    assert token_end is not None
    assert token_start <= token_end
    return (token_start, token_end + 1)


def batch_convert_ids_to_tokens(
    batch: Sequence[Sequence[int]], tokenizer: Tokenizer, **kwargs: Any
) -> Sequence[StrSequence]:
    """Call `convert_ids_to_tokens` on every sequence in the batch."""
    return [tokenizer.convert_ids_to_tokens(ids, **kwargs) for ids in batch]


@contextmanager
def set_padding_side(
    tokenizer: Tokenizer, padding_side: str = "right"
) -> Iterator[None]:
    """Temporarily set padding side for tokenizer.

    Useful for when you want to batch generate with causal LMs like GPT, as these
    require the padding to be on the left side in such settings but are much easier
    to mess around with when the padding, by default, is on the right.

    Example usage:
        tokenizer = transformers.AutoTokenizer.from_pretrained("gpt2")
        with tokenizer_utils.set_padding_side(tokenizer, "left"):
            inputs = mt.tokenizer(...)
        # ... later
        model.generate(**inputs)

    """
    _padding_side = tokenizer.padding_side
    tokenizer.padding_side = padding_side
    yield
    tokenizer.padding_side = _padding_side



"""Heuristics and tools for cleaning language."""
import logging
from functools import cache

from nltk.corpus import cmudict

logger = logging.getLogger(__name__)


@cache
def _cmudict() -> dict[str, list]:
    """Read the nltk cmudict corpus."""
    return cmudict.dict()


def determine_article(word: str, default: str = "a") -> str:
    """Determine article that should precede text."""
    pronounciation = _cmudict().get(word)
    if not pronounciation:
        logger.debug(f"no pronounciation found for: {word}")
        return default
    first = pronounciation[0]
    while not isinstance(first, str) and first:
        first = first[0]
    is_vowel_sound = first[-1].isdigit()
    return "an" if is_vowel_sound else "a"



"""Logic for getting and mucking with model hidden representations."""
import argparse
from functools import partial
from typing import Any, Literal, Optional, Sequence, cast, overload

from evaluation import data, models
from evaluation.utils import tokenizer_utils
from evaluation.utils.typing import (
    Dataset,
    Device,
    ModelInput,
    ModelOutput,
    StrSequence,
    Tokenizer,
    TokenizerOffsetMapping,
)

import torch
from baukit import nethook


def _remove_sent_case(text: str) -> str:
    """Make the string NOT sentence case (first letter lowercase)."""
    return text[0].lower() + text[1:]


def _is_batched(text: str | StrSequence) -> bool:
    """Determine if text is batched or not."""
    return not isinstance(text, str)


def _maybe_batch(text: str | StrSequence) -> StrSequence:
    """Batch the text if it is not already batched."""
    if isinstance(text, str):
        return [text]
    return text


def _as_fp32(data: dict) -> dict:
    """Cast all top-level float tensor values to float32."""
    return {
        key: value.float()
        if isinstance(value, torch.Tensor) and value.dtype.is_floating_point
        else value
        for key, value in data.items()
    }


def _validate_lengths(lengths: torch.Tensor) -> None:
    """Validate sequence lengths tensor is correct shape."""
    if len(lengths.shape) != 1:
        raise ValueError(f"misshapen lengths: {lengths.shape}")


def _validate_token_ranges(
    token_ranges: torch.Tensor, batch_size: int | None = None
) -> None:
    """Validate token ranges are correct shape."""
    if len(token_ranges.shape) != 2 or token_ranges.shape[1] != 2:
        raise ValueError(f"misshapen token ranges: {token_ranges.shape}")
    if batch_size is not None and token_ranges.shape[0] != batch_size:
        raise ValueError(
            f"expected batch_size={batch_size}, got {token_ranges.shape[0]}"
        )


def inputs_from_batch(
    mt: models.ModelAndTokenizer,
    text: str | StrSequence,
    device: Optional[Device] = None,
) -> tuple[ModelInput, Sequence[TokenizerOffsetMapping]]:
    """Precompute model inputs."""
    inputs = mt.tokenizer(
        text,
        return_tensors="pt",
        truncation=True,
        padding="longest",
        return_offsets_mapping=True,
    )
    offset_mapping = inputs.pop("offset_mapping")
    if device is not None:
        inputs = inputs.to(device)
    return inputs, offset_mapping


def last_token_index_from_batch(inputs: ModelInput) -> Sequence[int]:
    """Return index of last token for each item in batch, accounting for padding."""
    return inputs.attention_mask.sum(dim=-1).cpu() - 1


HiddensByLayer = dict[int, torch.Tensor]


@overload
def hiddens_from_batch(
    mt: models.ModelAndTokenizer,
    inputs: str | StrSequence | ModelInput,
    stop: Literal[True] = True,
    layers: Optional[Sequence[int]] = None,
    device: Optional[Device] = None,
) -> HiddensByLayer:
    ...


@overload
def hiddens_from_batch(
    mt: models.ModelAndTokenizer,
    inputs: str | StrSequence | ModelInput,
    stop: Literal[False],
    layers: Optional[Sequence[int]] = None,
    device: Optional[Device] = None,
) -> tuple[HiddensByLayer, ModelOutput]:
    ...


@torch.inference_mode()
def hiddens_from_batch(
    mt: models.ModelAndTokenizer,
    inputs: str | StrSequence | ModelInput,
    stop: bool = True,
    layers: Optional[Sequence[int]] = None,
    device: Optional[Device] = None,
) -> HiddensByLayer | tuple[HiddensByLayer, ModelOutput]:
    """Precomptue hidden reps.

    Args:
        mt: The model and tokenizer.
        inputs: The model inputs.
        stop: Stop computation after retrieving the hiddens.
        layers: Layers to compute hiddens for. Defaults to all.
        device: Send model and inputs to this device.

    Returns:
        Hidden reps mapped by layer and model output (if stop=False).

    """
    mt.to_(device)
    if isinstance(inputs, str | list | tuple):
        inputs, _ = inputs_from_batch(mt, inputs, device=device)
    if device is not None:
        inputs = inputs.to(device)
    outputs = None
    layer_paths = models.determine_layer_paths(mt, layers=layers, return_dict=True)
    with nethook.TraceDict(mt.model, layers=layer_paths.values(), stop=stop) as ret:
        outputs = mt.model(**inputs)

    hiddens_by_layer = {}
    for layer, layer_path in layer_paths.items():
        hiddens = ret[layer_path].output
        if isinstance(hiddens, tuple):
            hiddens = hiddens[0]
        assert isinstance(hiddens, torch.Tensor)
        hiddens_by_layer[layer] = hiddens.cpu()

    return hiddens_by_layer if stop else (hiddens_by_layer, outputs)


def token_ranges_from_batch(
    strings: str | StrSequence,
    substrings: str | StrSequence,
    offsets_mapping: Sequence[TokenizerOffsetMapping],
    occurrence: int = 0,
) -> torch.Tensor:
    """Return shape (batch_size, 2) tensor of token ranges for (str, substr) pairs."""
    strings = _maybe_batch(strings)
    substrings = _maybe_batch(substrings)
    if len(strings) != len(substrings):
        raise ValueError(
            f"got {len(strings)} strings but only {len(substrings)} substrings"
        )

    return torch.tensor(
        [
            tokenizer_utils.find_token_range(
                string, substring, offset_mapping=offset_mapping, occurrence=occurrence
            )
            for string, substring, offset_mapping in zip(
                strings, substrings, offsets_mapping
            )
        ]
    )


def last_token_ranges_from_batch(token_ranges: torch.Tensor) -> torch.Tensor:
    """Convert batch of token ranges to only include last token."""
    _validate_token_ranges(token_ranges)
    token_ranges = token_ranges.clone()
    token_ranges[:, 0] = token_ranges[:, 1] - 1
    return token_ranges


def negative_token_ranges_from_batch(
    token_ranges: torch.Tensor, lengths: torch.Tensor
) -> torch.Tensor:
    """Convert positive token ranges to negative ones."""
    _validate_lengths(lengths)
    _validate_token_ranges(token_ranges, batch_size=len(lengths))
    return token_ranges.cpu() - lengths[:, None].cpu()


def first_token_ids_from_batch(
    mt: models.ModelAndTokenizer | Tokenizer, words: str | StrSequence, add_space: bool = True,
) -> torch.Tensor:
    """Return shape (batch_size,) int tensor with first token ID for each word."""
    words = _maybe_batch(words)
    tokenizer = models.unwrap_tokenizer(mt)
    # TODO(evandez): Centralize this spacing nonsense.
    if add_space:
        token_ids = tokenizer([" " + word for word in words], add_special_tokens=False)
    else:
        token_ids = tokenizer(words, add_special_tokens=False)
    return torch.tensor([ti[0] for ti in token_ids.input_ids])


def average_hiddens_from_batch(
    hiddens: torch.Tensor, ranges: Sequence[Sequence[int]] | torch.Tensor
) -> torch.Tensor:
    """Compute average hidden rep in given token ranges.

    Args:
        hiddens: Should have shape (batch_size, sequence_length, hidden_size)
        ranges: Token ranges.

    Returns:
        Shape (batch_size, hidden_size) tensor containing average hiddens.

    """
    if not isinstance(ranges, torch.Tensor):
        ranges = torch.tensor(ranges)
    if ranges.shape != (hiddens.shape[0], 2):
        raise ValueError(f"unexpected ranges shape: {ranges.shape}")

    averages = []
    for bi, (ti, tj) in enumerate(ranges.tolist()):
        average = hiddens[bi, ti:tj].mean(dim=0)
        averages.append(average)

    return torch.stack(averages)


def editor_inputs_from_batch(
    mt: models.ModelAndTokenizer,
    batch: data.ContextMediationInput,
    layers: Optional[Sequence[int]] = None,
    device: Optional[Device] = None,
    entity_occurrence_in_prompt: int = 0,
    return_token_ranges: bool = False,
    return_target_token_ids: bool = True,
    return_entity_hiddens: bool = False,
    return_attribute_hiddens: bool = False,
    fp32: bool = False,
    target_token_first_space: bool = True, 
) -> dict:
    """Precompute everything the editor model needs to run from the batch."""
    mt.model.to(device)

    # Pull out expected values.
    entities = _maybe_batch(batch["entity"])
    prompts = _maybe_batch(batch["prompt"])
    contexts = _maybe_batch(batch["context"])
    attributes = _maybe_batch(batch["attribute"])

    precomputed: dict = {}

    # Precompute inputs.
    prompt_inputs, prompt_offset_mapping = None, None
    context_inputs, context_offset_mapping = None, None
    entity_inputs, entity_offset_mapping = None, None
    with models.set_padding_side(mt, padding_side="right"):
        if return_token_ranges:
            prompt_inputs, prompt_offset_mapping = inputs_from_batch(mt, prompts)
        if return_attribute_hiddens or return_token_ranges:
            context_inputs, context_offset_mapping = inputs_from_batch(
                mt, contexts, device=device
            )
        if return_entity_hiddens or return_token_ranges:
            entity_inputs, entity_offset_mapping = inputs_from_batch(
                mt, entities, device=device
            )

    # Precompute token ranges if needed.
    if return_token_ranges or return_attribute_hiddens:
        assert prompt_inputs is not None and prompt_offset_mapping is not None
        assert context_inputs is not None and context_offset_mapping is not None
        assert entity_inputs is not None and entity_offset_mapping is not None
        for (
            key_string,
            key_substring,
            strings,
            substrings,
            inputs,
            offset_mapping,
            occurrence,
        ) in (
            (
                "entity",
                "entity",
                entities,
                entities,
                entity_inputs,
                entity_offset_mapping,
                0,
            ),
            (
                "prompt",
                "entity",
                prompts,
                entities,
                prompt_inputs,
                prompt_offset_mapping,
                entity_occurrence_in_prompt,
            ),
            (
                "context",
                "attribute",
                contexts,
                attributes,
                context_inputs,
                context_offset_mapping,
                0,
            ),
        ):
            lengths = inputs.attention_mask.sum(dim=-1).cpu()
            precomputed[f"{key_string}.length"] = lengths

            key = f"{key_string}.{key_substring}"

            key_tr_base = f"{key}.token_range"
            precomputed[key_tr_base] = tr = token_ranges_from_batch(
                strings,
                substrings,
                offsets_mapping=offset_mapping,
                occurrence=occurrence,
            )

            key_tr_neg = f"{key}.negative_token_range"
            precomputed[key_tr_neg] = negative_token_ranges_from_batch(tr, lengths)

            key_tr_base_last = f"{key_tr_base}.last"
            precomputed[key_tr_base_last] = ltr = last_token_ranges_from_batch(tr)

            key_tr_neg_last = f"{key_tr_neg}.last"
            precomputed[key_tr_neg_last] = negative_token_ranges_from_batch(
                ltr, lengths
            )

    # Precompute token IDs if needed.
    if return_target_token_ids:
        for target_key in ("target_mediated", "target_unmediated"):
            target = cast(str | list, batch.get(target_key))
            if target is None or any(t is None for t in target):
                continue
            precomputed[f"{target_key}.token_id"] = first_token_ids_from_batch(
                mt, target, target_token_first_space 
            )

    # Precompute average/last hidden reps if needed.
    for (
        key_string,
        key_substring,
        condition,
        inputs,
    ) in (
        ("entity", "entity", return_entity_hiddens, entity_inputs),
        ("context", "attribute", return_attribute_hiddens, context_inputs),
    ):
        if not condition:
            continue

        key = f"{key_string}.{key_substring}"

        key_token_range = f"{key}.token_range"
        token_ranges = precomputed[key_token_range]
        token_ranges_last = precomputed[f"{key_token_range}.last"]

        hiddens_by_layer = hiddens_from_batch(mt, inputs, layers=layers, device=device)
        for layer, hiddens in hiddens_by_layer.items():
            key_hiddens = f"{key}.hiddens.{layer}"
            key_hiddens_last = f"{key_hiddens}.last"
            precomputed[key_hiddens_last] = average_hiddens_from_batch(
                hiddens, token_ranges_last
            )

            key_hiddens_average = f"{key_hiddens}.average"
            precomputed[key_hiddens_average] = average_hiddens_from_batch(
                hiddens, token_ranges
            )

    if fp32:
        precomputed = _as_fp32(precomputed)

    return precomputed


def edit_input_string_attn_mask(
        strings: list, 
        substrings: list, 
        inputs: ModelInput,
        offset_mapping,
        attn_scale: float, 
        attn_scale_pos: str, 
        occurrence: int = 0,
        device: Optional[Device] = None,
    ):
    # Calculate the token ranges for attribute
    precomputed: dict = dict()
    key_tr_base = f"token_range"
    precomputed[key_tr_base] = tr = token_ranges_from_batch(
        strings,
        substrings,
        offsets_mapping=offset_mapping,
        occurrence=occurrence,
    )

    attention_mask = inputs['attention_mask'].float().clone()
    for bi, (ti, tj) in enumerate(tr.tolist()):
        if attn_scale_pos == "include":
            attention_mask[bi, ti:tj] = attention_mask[bi, ti:tj]*attn_scale
        elif attn_scale_pos == "exclude":
            attention_mask[bi, :ti] = attention_mask[bi, :ti]*attn_scale
            attention_mask[bi, tj:] = attention_mask[bi, tj:]*attn_scale
        else:
            raise ValueError(f"Unimplemented scale pos: {str(attn_scale_pos)}")
    inputs['attention_mask'] = attention_mask
        
    if device is not None:
        inputs = inputs.to(device)
    
    return inputs, precomputed


def edit_input_attn_mask_for_biosbias(
        batch: data.ContextMediationInput, 
        key_string: str,
        key_substring: str,
        inputs: ModelInput,
        offset_mapping,
        attn_scale: float, 
        attn_scale_pos: str, 
        occurrence: int = 0,
        device: Optional[Device] = None,
    ):
    strings = _maybe_batch(batch[key_string]) 
    substrings = _maybe_batch(batch[key_substring]) 
    precomputed: dict = dict()

    # Calculate the token ranges for attribute
    key = f"{key_string}.{key_substring}"
    key_tr_base = f"{key}.token_range"
    precomputed[key_tr_base] = tr = token_ranges_from_batch(
        strings,
        substrings,
        offsets_mapping=offset_mapping,
        occurrence=occurrence,
    )

    attention_mask = inputs['attention_mask'].float().clone()
    for bi, (ti, tj) in enumerate(tr.tolist()):
        if attn_scale_pos == "include":
            attention_mask[bi, ti:tj] = attention_mask[bi, ti:tj]*attn_scale
        elif attn_scale_pos == "exclude":
            attention_mask[bi, :ti] = attention_mask[bi, :ti]*attn_scale
            attention_mask[bi, tj:] = attention_mask[bi, tj:]*attn_scale
        else:
            raise ValueError(f"Unimplemented scale pos: {str(attn_scale_pos)}")
    inputs['attention_mask'] = attention_mask
        
    if device is not None:
        inputs = inputs.to(device)
    
    return inputs, precomputed


def edit_input_attn_mask_from_batch(
    mt: models.ModelAndTokenizer,
    batch: data.ContextMediationInput,
    attn_scale: float | None = None, 
    attn_scale_pos: str | None = None,  
    add_unmediated_fact: str|None = None, 
    add_marker: str|None = None,
    fewshot_examples: str|None = None, 
    device: Optional[Device] = None,
) -> dict:
    """Precompute everything the editor model needs to run from the batch."""
    mt.model.to(device)

    # Pull out expected values.
    prompts = _maybe_batch(batch["prompt"])
    contexts = _maybe_batch(batch["context"])
    attributes = _maybe_batch(batch["attribute"])
    targets_mediated = _maybe_batch(batch["target_mediated"])
    targets_unmediated = _maybe_batch(batch["target_unmediated"])

    if add_unmediated_fact is not None:
        new_prompts = []
        for prompt, context, target_mediated, target_unmediated in zip(
            prompts, contexts, targets_mediated, targets_unmediated
        ):
            unmediated_prefix = "Previously "
            mediated_prefix = "Currently "
            if add_unmediated_fact == "left":
                unmediated_fact = context.replace(target_mediated, target_unmediated)+". "
                new_prompt = f"{unmediated_prefix}{unmediated_fact}{mediated_prefix}{prompt}"
            elif add_unmediated_fact == "right":
                prompt = prompt.replace(target_mediated, target_unmediated)
                context = context.rstrip(". ") + ". "
                new_prompt = f"{mediated_prefix}{context}{unmediated_prefix}{prompt}"
            new_prompts.append(new_prompt)

        prompts = new_prompts

        if add_marker is not None:
            prompts = [
                prompt.replace(attr, add_marker+attr+add_marker) for prompt,attr in zip(prompts, attributes)
            ]
        
        if fewshot_examples is not None:
            prompts = [
                fewshot_examples + prompt for prompt in prompts
            ]

    # Precompute inputs.
    prompt_inputs, prompt_offset_mapping = None, None
    with models.set_padding_side(mt, padding_side="left"):
        prompt_inputs, prompt_offset_mapping = inputs_from_batch(mt, prompts)

    # precomputed: dict = {}
      
    def edit_attention_mask(
        key_string,
        key_substring,
        strings,
        substrings,
        inputs,
        attn_scale, 
        attn_scale_pos,
        offset_mapping,
        occurrence,
    ):
        precomputed: dict = {}
        lengths = inputs.attention_mask.sum(dim=-1).cpu()
        precomputed[f"{key_string}.length"] = lengths

        key = f"{key_string}.{key_substring}"
        key_tr_base = f"{key}.token_range"
        precomputed[key_tr_base] = tr = token_ranges_from_batch(
            strings,
            substrings,
            offsets_mapping=offset_mapping,
            occurrence=occurrence,
        )

        if isinstance(attn_scale, float):
            attention_mask = inputs['attention_mask'].float().clone()
            for bi, (ti, tj) in enumerate(tr.tolist()):
                if attn_scale_pos == "include":
                    attention_mask[bi, ti:tj] = attention_mask[bi, ti:tj]*attn_scale
                elif attn_scale_pos == "exclude":
                    attention_mask[bi, :ti] = attention_mask[bi, :ti]*attn_scale
                    attention_mask[bi, tj:] = attention_mask[bi, tj:]*attn_scale
                else:
                    raise ValueError(f"Unimplemented scale pos: {str(attn_scale_pos)}")
            inputs['attention_mask'] = attention_mask
            
        if device is not None:
            inputs = inputs.to(device)
        
        return inputs, precomputed
    
    new_inputs, precomputed = edit_attention_mask(
            "prompt",
            "attribute",
            prompts,
            attributes,
            prompt_inputs,
            attn_scale, 
            attn_scale_pos,
            prompt_offset_mapping,
            0,
        )
    precomputed["new_prompt"] = prompts 

    return new_inputs, precomputed


def prepare_counterfact_few_shot_examples(
        fewshot_samples, 
        add_unmediated_fact="left", unmediated_prefix="Previously ", mediated_prefix="Currently ",
        example_sep = "\n###\n", text_target_sep = " "
    ):
    labeled_examples = [] 
    for example in fewshot_samples:
        prompt = example['prompt']
        context = example['context']
        target_mediated = example['target_mediated']
        target_unmediated = example['target_unmediated']

        if add_unmediated_fact == "left":
            unmediated_fact = context.replace(target_mediated, target_unmediated)+". "
            new_prompt = f"{unmediated_prefix}{unmediated_fact}{mediated_prefix}{prompt}"
        elif add_unmediated_fact == "right":
            prompt = prompt.replace(target_mediated, target_unmediated)
            context = context.rstrip(". ") + ". "
            new_prompt = f"{mediated_prefix}{context}{unmediated_prefix}{prompt}"
        else:
            raise ValueError("Unexpected %s"%(add_unmediated_fact))

        # text = example[text_key]
        # entity = example['entity']
        # target = example['target_mediated'] 
        # text_label = text + text_target_sep + target 

        text_label = new_prompt + text_target_sep + target_mediated
        labeled_examples.append(text_label) 
    
    fewshot_examples = example_sep.join(labeled_examples) + example_sep
    return fewshot_examples 


def editor_inputs_from_dataset(
    mt: models.ModelAndTokenizer,
    dataset: Dataset,
    layers: Optional[Sequence[int]] = None,
    device: Optional[Device] = None,
    batch_size: int = 64,
    desc: str | None = "precompute editor inputs",
    **kwargs: Any,
) -> Dataset:
    """Precompute everything the editor model needs to train and run."""
    if "fp32" in kwargs:
        raise ValueError("cannot set fp32= because arrow datasets only support fp32")
    return dataset.map(
        partial(
            editor_inputs_from_batch,
            mt,
            layers=layers,
            device=device,
            fp32=True,
            **kwargs,
        ),
        batched=True,
        batch_size=batch_size,
        desc=desc,
        keep_in_memory=True,
        num_proc=1,
    )


def has_editor_inputs(batch: dict) -> bool:
    """Determine if editor inputs already precomputed."""
    return "prompt.entity.token_range" in batch  # Check for just one flag entry.


def prompt_in_context_from_sample(
    entity: str,
    prompt: str,
    context: str,
    context_prefix: str | None = None,
    context_suffix: str | None = None,
    prompt_prefix: str | None = None,
) -> str:
    """Compute prompt in context for the sample.

    The prompt in context is simply the "prompt" field in each sample prepended with
    the "context" field. This function tries to make the casing look sensible while
    also not changing the casing of any entity mention.

    Can optionally include prefixes for all contexts and/or for all prompts. This is
    useful for adding function or transition words between the prompt and context so
    that the language model can better reconcile the task.

    Args:
        entity: The entity.
        prompt: The prompt.
        context: The context.
        context_prefix: Prepend this to context.
        context_suffix: Append this to context.
        prompt_prefix: Prepend this to prompt, but after context.


    Returns:
        A single string with the context followed by the prompt.

    """
    if prompt_prefix is not None:
        if not prompt.startswith(entity):
            prompt = _remove_sent_case(prompt)
        prompt = f"{prompt_prefix}{prompt}"

    if context_prefix is not None:
        if not context.startswith(entity):
            context = _remove_sent_case(context)
        context = f"{context_prefix} {context}"

    # Always make sure context is a complete, period-ended sentence.
    context = context.rstrip(". ") + "."

    if context_suffix is not None:
        context = f"{context}{context_suffix}"
    else:
        context += " "

    prompt_in_context = f"{context}{prompt}"
    return prompt_in_context


def prompt_in_context_from_batch(
    batch: data.ContextMediationInput,
    output_key: str = "prompt_in_context",
    **kwargs: Any,
) -> dict:
    """Compute prompt in context from batch."""
    is_batched = _is_batched(batch["entity"])
    entities = _maybe_batch(batch["entity"])
    prompts = _maybe_batch(batch["prompt"])
    contexts = _maybe_batch(batch["context"])

    prompts_in_context = []
    for entity, prompt, context in zip(entities, prompts, contexts):
        prompt_in_context = prompt_in_context_from_sample(
            entity, prompt, context, **kwargs
        )
        prompts_in_context.append(prompt_in_context)

    return {output_key: prompts_in_context if is_batched else prompts_in_context[0]}


def prompt_in_context_from_dataset(
    dataset: Dataset, desc: str | None = "precompute prompt in context", **kwargs: Any
) -> Dataset:
    """Compute prompt in context for whole dataset."""
    return dataset.map(
        partial(prompt_in_context_from_batch, **kwargs),
        desc=desc,
        keep_in_memory=True,
    )


def has_prompt_in_context(batch: dict) -> bool:
    """Check if prompt_in_context has already been computed."""
    return "prompt_in_context" in batch


def entity_deltas_from_batch(
    mt: models.ModelAndTokenizer,
    batch: data.ContextMediationInput,
    layers: Optional[Sequence[int]] = None,
    device: Optional[Device] = None,
    fp32: bool = False,
    return_token_ranges: bool = True,
    return_deltas: bool = True,
    **kwargs: Any,
) -> dict:
    """Compute in-context entity delta for the batch.

    Keyword arguments are forwarded to `prompt_in_context_from_batch`.

    Args:
        mt: Model and tokenizer.
        batch: Context mediation-style batch.
        layers: Layers to compute deltas in. Defaults to all.
        device: Send model and inputs to this device.
        fp32: Force cast each tensor to fp32.
        return_token_ranges: Return entity token ranges for contextualied prompt.
        return_deltas: Return entity deltas for each layer.

    Returns:
        Deltas and related precomputed values.

    """
    precomputed: dict = prompt_in_context_from_batch(batch, **kwargs)

    entities = _maybe_batch(batch["entity"])
    prompts_in_context = precomputed["prompt_in_context"]

    inputs = None
    first_entity_token_ranges = None
    last_entity_token_ranges = None
    if return_token_ranges or return_deltas:
        with models.set_padding_side(mt, padding_side="right"):
            inputs, offset_mapping = inputs_from_batch(
                mt, prompts_in_context, device=device
            )
        first_entity_token_ranges = token_ranges_from_batch(
            prompts_in_context, entities, offset_mapping
        )
        last_entity_token_ranges = token_ranges_from_batch(
            prompts_in_context, entities, offset_mapping, occurrence=1
        )

    if return_token_ranges:
        assert first_entity_token_ranges is not None
        assert last_entity_token_ranges is not None
        for position, token_ranges in (
            ("first", first_entity_token_ranges),
            ("last", last_entity_token_ranges),
        ):
            key = f"prompt_in_context.entity.token_range.{position}"
            precomputed[key] = token_ranges

    if return_deltas:
        assert inputs is not None
        assert first_entity_token_ranges is not None
        assert last_entity_token_ranges is not None
        hiddens_by_layer = hiddens_from_batch(mt, inputs, layers=layers, device=device)
        for layer, hiddens in hiddens_by_layer.items():
            first_entity_hiddens = average_hiddens_from_batch(
                hiddens, first_entity_token_ranges
            )
            last_entity_hiddens = average_hiddens_from_batch(
                hiddens, last_entity_token_ranges
            )
            delta = last_entity_hiddens - first_entity_hiddens

            key = f"prompt_in_context.entity.delta.{layer}"
            precomputed[key] = delta

    if fp32:
        precomputed = _as_fp32(precomputed)

    return precomputed


def entity_deltas_from_dataset(
    mt: models.ModelAndTokenizer,
    dataset: Dataset,
    layers: Optional[Sequence[int]] = None,
    device: Optional[Device] = None,
    batch_size: int = 64,
    desc: str | None = "precompute entity deltas",
    **kwargs: Any,
) -> Dataset:
    """Precompute entity deltas in context for the whole dataset."""
    return dataset.map(
        partial(
            entity_deltas_from_batch,
            mt,
            layers=layers,
            device=device,
            fp32=True,
            **kwargs,
        ),
        batched=True,
        batch_size=batch_size,
        desc=desc,
        keep_in_memory=True,
        num_proc=1,
    )


def has_entity_deltas(batch: dict) -> bool:
    """Return True if the batch already has precomputed entity deltas."""
    return any("delta" in key for key in batch)


def classification_inputs_from_batch(
    mt: models.ModelAndTokenizer,
    batch: data.ContextMediationInput,
    layers: Optional[Sequence[int]] = None,
    device: Optional[Device] = None,
    fp32: bool = False,
    **kwargs: Any,
) -> dict:
    """Precompute classification inputs for the batch.

    An extension of `editor_inputs_from_batch` that additionally computes attribute
    directions for the unmediated case and entity representations in context.

    Args:
        mt: Model and tokenizer.
        batch: The batch.
        layers: Model layers to classify entities at.
        device: Send model and inputs to this device.
        fp32: Force cast each tensor to fp32.

    Returns:
        Batch with precomputed values.

    """
    precomputed: dict = {**batch}
    if not has_editor_inputs(precomputed):
        precomputed = editor_inputs_from_batch(
            mt, batch, layers=layers, device=device, fp32=fp32, **kwargs
        )
    if not has_prompt_in_context(precomputed):
        precomputed.update(prompt_in_context_from_batch(batch))

    entities = _maybe_batch(batch["entity"])
    prompts = _maybe_batch(batch["prompt"])
    prompts_in_context = cast(list[str], precomputed["prompt_in_context"])

    contexts_m = _maybe_batch(batch["context"])
    attributes_m = _maybe_batch(batch["attribute"])

    targets_m = batch["target_mediated"]
    if targets_m is None:
        raise ValueError("target_mediated cannot be None")
    targets_m = _maybe_batch(targets_m)
    targets_m_ids = precomputed["target_mediated.token_id"]

    targets_u = batch.get("target_unmediated")
    targets_u_ids = None
    contexts_u = None
    attributes_u = None
    if targets_u is not None and all(t is not None for t in targets_u):
        targets_u = _maybe_batch(targets_u)
        targets_u_ids = precomputed["target_unmediated.token_id"]

        precomputed["context_unmediated"] = contexts_u = [
            context.replace(target_m, target_u)
            for context, target_m, target_u in zip(contexts_m, targets_m, targets_u)
        ]
        precomputed["attribute_unmediated"] = attributes_u = [
            attribute.replace(target_m, target_u)
            for attribute, target_m, target_u in zip(attributes_m, targets_m, targets_u)
        ]

    for (
        key_string,
        key_substring,
        strings,
        substrings,
        occurrence,
        target_ids,
        comparator_ids,
    ) in (
        (
            "context_unmediated",
            "attribute_unmediated",
            contexts_u,
            attributes_u,
            0,
            None,
            None,
        ),
        (
            "prompt_in_context",
            "entity",
            prompts_in_context,
            entities,
            1,
            targets_m_ids,
            targets_u_ids,
        ),
        (
            "prompt",
            "entity",
            prompts,
            entities,
            0,
            targets_u_ids,
            targets_m_ids,
        ),
    ):
        if strings is None or substrings is None:
            continue

        with models.set_padding_side(mt, padding_side="left"):
            inputs, offsets_mapping = inputs_from_batch(mt, strings, device=device)

        trs_all = token_ranges_from_batch(
            strings, substrings, offsets_mapping, occurrence=occurrence
        )
        trs_last = last_token_ranges_from_batch(trs_all)

        # NOTE(evandez): This is an odd place to do this, but since we've already done
        # the work of processing the prompt and prompt in context, we may as well
        # record the target probabilities.

        if target_ids is not None or comparator_ids is not None:
            hiddens_by_layer, outputs = hiddens_from_batch(
                mt, inputs, layers=layers, device=device, stop=False
            )
            log_probs = torch.log_softmax(outputs.logits[:, -1], dim=-1)
            batch_idx = torch.arange(len(strings))
            if target_ids is not None:
                precomputed[f"{key_string}.target.logp"] = log_probs[
                    batch_idx, target_ids
                ]
            if comparator_ids is not None:
                precomputed[f"{key_string}.comparator.logp"] = log_probs[
                    batch_idx, comparator_ids
                ]
        else:
            hiddens_by_layer = hiddens_from_batch(
                mt, inputs, layers=layers, device=device
            )

        for layer, hiddens in hiddens_by_layer.items():
            key = f"{key_string}.{key_substring}.hiddens.{layer}"
            precomputed[f"{key}.average"] = average_hiddens_from_batch(hiddens, trs_all)
            precomputed[f"{key}.last"] = average_hiddens_from_batch(hiddens, trs_last)

    if fp32:
        precomputed = _as_fp32(precomputed)

    return precomputed


def classification_inputs_from_dataset(
    mt: models.ModelAndTokenizer,
    dataset: Dataset,
    layers: Optional[Sequence[int]] = None,
    device: Optional[Device] = None,
    batch_size: int = 64,
    desc: str | None = "precompute classification inputs",
    **kwargs: Any,
) -> Dataset:
    """Precompute classification inputs for the whole dataset."""
    return dataset.map(
        partial(
            classification_inputs_from_batch,
            mt,
            layers=layers,
            device=device,
            fp32=True,
            **kwargs,
        ),
        batched=True,
        batch_size=batch_size,
        desc=desc,
        keep_in_memory=True,
        num_proc=1,
    )


def has_classification_inputs(batch: dict) -> bool:
    """Determine if batch already has precomputed classification inputs."""
    return "context_unmediated" in batch


PromptKey = Literal["prompt", "prompt_in_context"]
TargetKey = Literal["target_mediated", "target_unmediated"]


@torch.inference_mode()
def model_predictions_from_batch(
    mt: models.ModelAndTokenizer,
    batch: dict,
    device: Device | None = None,
    return_top_k: int = 5,
    input_prompt_key: PromptKey = "prompt",
    input_target_key: TargetKey | None = "target_unmediated",
    input_comparator_key: TargetKey | None = "target_mediated",
    other_targets: StrSequence | None = None,
    other_targets_idx: Sequence[int] | None = None,
    output_correct_key: str = "model_correct",
    output_other_targets_key: str = "other_targets",
    output_top_tokens_key: str = "top_tokens",
) -> dict:
    """Precompute model predictions on prompt from the batch."""
    if other_targets is not None and other_targets_idx is None:
        other_targets_idx = first_token_ids_from_batch(mt, other_targets).tolist()

    prompts = batch[input_prompt_key]
    with models.set_padding_side(mt, padding_side="left"):
        inputs, _ = inputs_from_batch(mt, prompts, device=device)
    outputs = mt.model(**inputs)
    distribution = torch.log_softmax(outputs.logits[:, -1], dim=-1)

    precomputed = {}

    top_tokens_idx = distribution.topk(k=return_top_k, dim=-1).indices.tolist()
    top_tokens = tokenizer_utils.batch_convert_ids_to_tokens(
        top_tokens_idx, mt.tokenizer
    )
    precomputed[f"{input_prompt_key}.{output_top_tokens_key}"] = top_tokens

    batch_idx = torch.arange(len(prompts))
    if input_target_key:
        targets = batch[input_target_key]
        targets_token_idx = first_token_ids_from_batch(mt, targets)
        targets_log_p = distribution[batch_idx, targets_token_idx]
        precomputed[
            f"{input_prompt_key}.{input_target_key}.logp"
        ] = targets_log_p.tolist()

    if input_comparator_key is not None:
        comparators = batch[input_comparator_key]
        comparators_token_idx = first_token_ids_from_batch(mt, comparators)
        comparators_log_p = distribution[batch_idx, comparators_token_idx]
        precomputed[
            f"{input_prompt_key}.{input_comparator_key}.logp"
        ] = comparators_log_p.tolist()

    if other_targets_idx is not None:
        key = f"{input_prompt_key}.{output_other_targets_key}.logp"
        precomputed[key] = distribution[:, other_targets_idx].tolist()

    if input_target_key is not None and input_comparator_key is not None:
        precomputed[f"{input_prompt_key}.{output_correct_key}"] = targets_log_p.gt(
            comparators_log_p
        ).tolist()

    return precomputed


def model_predictions_from_dataset(
    mt: models.ModelAndTokenizer,
    dataset: Dataset,
    device: Optional[Device] = None,
    batch_size: int = 64,
    desc: str | None = "precompute model predictions",
    other_targets: StrSequence | None = None,
    other_targets_idx: Sequence[int] | None = None,
    **kwargs: Any,
) -> Dataset:
    """Precompute model predictions for the whole dataset."""
    if other_targets is not None and other_targets_idx is None:
        other_targets_idx = first_token_ids_from_batch(mt, other_targets).tolist()
    return dataset.map(
        partial(
            model_predictions_from_batch,
            mt,
            device=device,
            other_targets_idx=other_targets_idx,
            **kwargs,
        ),
        batched=True,
        batch_size=batch_size,
        desc=desc,
        keep_in_memory=True,
        num_proc=1,
    )


def add_preprocessing_args(parser: argparse.ArgumentParser) -> None:
    """Add common preprocessing args.

    The args include:
        --attribute-no-entity: When computing attribute rep, do not use entity.

    """
    parser.add_argument(
        "--attribute-no-entity",
        action="store_true",
        default=False,
        help="set context = attribute",
    )


def from_args(args: argparse.Namespace, dataset: Dataset) -> Dataset:
    """Apply all the preprocessing steps determined by the args."""
    if args.attribute_no_entity:
        dataset = dataset.map(
            lambda e: {"context": e["attribute"]}, desc="set context=attribute"
        )
    return dataset



"""Some useful type aliases relevant to this project."""
import pathlib
from typing import Sequence

import datasets
import numpy
import torch
import transformers
import transformers.modeling_outputs

ArrayLike = list | tuple | numpy.ndarray | torch.Tensor
PathLike = str | pathlib.Path
Device = str | torch.device

# Throughout this codebase, we use HuggingFace model implementations
# as well as HuggingFace datasets.
Model = (
    transformers.GPTJForCausalLM
    | transformers.LlamaForCausalLM
    | transformers.GPT2LMHeadModel
    | transformers.GPTNeoXForCausalLM
)
Tokenizer = transformers.PreTrainedTokenizerFast
TokenizerOffsetMapping = Sequence[tuple[int, int]]
Dataset = datasets.arrow_dataset.Dataset | datasets.dataset_dict.DatasetDict
ModelInput = transformers.BatchEncoding
ModelOutput = transformers.modeling_outputs.CausalLMOutput
ModelGenerateOutput = transformers.generation.utils.GenerateOutput | torch.LongTensor

# All strings are also Sequence[str], so we have to distinguish that we
# mean lists or tuples of strings, or sets of strings, not other strings.
StrSequence = list[str] | tuple[str, ...]
